unit KozyTetris;
{ --------------------------------------------------
  TKozyTetris component by Gasper Kozak.
  E-mail: gasper@hsp.si, kozax@usa.net

  See "KozyTetris Readme.Txt" for more details.

  Component originally created on 20th of September,
  1999.

  Feel free to change/distribute/rape this code.
  Just leave this note here.
  -------------------------------------------------- }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

const
  MaxMatrixSize = 5;
  MaxPlayColumns = 30;
  MaxPlayRows    = 50;

type
  TMatrix = array[1..MaxMatrixSize, 1..MaxMatrixSize] of Boolean;
  TPlayMatrix = array[1..MaxPlayColumns, 1..MaxPlayRows] of Integer;

  TOnLineFull = procedure(Sender: TObject; Lines: Integer) of object;
  TOnGameEnd  = procedure(Sender: TObject) of object;

  TPiece = class
    PieceName: String;
    PieceColor: TColor;
    PieceSize: Integer;
    Matrix: TMatrix;
    Rot: TPoint;
    procedure Assign(Source: TPiece);
  end;

  TKozyTetris = class(TGraphicControl)
  private
    { Private declarations }
    pX, pY: Integer;
    FGridColor, FColor: TColor;
    FBlockSize, FRows, FColumns: Byte;
    FCanPlay, FPlaying, FDrawGrid: Boolean;
    FPieceList: TList;
    FCurrentPiece: TPiece;
    FEmptyLines, FCurrentPieceIndex: Integer;
    Field: TPlayMatrix;

    FOnLineFull: TOnLineFull;
    FOnGameEnd: TOnGameEnd;
  protected
    { Protected declarations }
    FPieces: TStringList;

    procedure WriteBlockSize(NewBlockSize: Byte);
    procedure WriteRows(NewRows: Byte);
    procedure WriteColumns(NewColumns: Byte);
    procedure WriteColor(NewColor: TColor);
    procedure WriteGridColor(NewGridColor: TColor);
    procedure WriteDrawGrid(NewDrawGrid: Boolean);

    procedure SetCurrentPieceIndex(Index: Integer);

    function ReadPieces: TStringList;
    procedure WritePieces(NewPieces: TStringList);

    function MapPieceToMatrix: Boolean;
    function SetNewPiece: integer;

    function InitializePieces: Boolean;
    procedure ClearPieceList;
    procedure ClearPlayField;

    procedure DrawPieceTo(Canvas: TCanvas; X, Y: Integer; Piece: TPiece; Erase: Boolean);
    procedure DrawPiece(X, Y: Integer; Piece: TPiece);
    procedure ErasePiece(X, Y: Integer; Piece: TPiece);
    procedure DrawCurrentPiece(X, Y: Integer);
    procedure EraseCurrentPiece(X, Y: Integer);
    function CanMove(X, Y: Integer): Boolean;
    function CanRotate(Clockwise: Boolean): Boolean;
    procedure CheckFullLines;
    function CountEmptyLines: Integer;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Paint; Override;

    procedure StartNewGame;

    function PieceCount: Integer;

    procedure RotatePiece(Clockwise: Boolean);
    function MoveLeft : Integer;
    function MoveRight : Integer;
    function MoveDown : Integer;
    function DropDown : Integer;

    property EmptyLines: Integer read CountEmptyLines;
    property Playing: Boolean read FPlaying;
    property CurrentPiece: Integer read FCurrentPieceIndex write SetCurrentPieceIndex;
  published
    { Published declarations }
    property Color: TColor read FColor write WriteColor;
    property GridColor: TColor read FGridColor write WriteGridColor;
    property BlockSize: Byte read FBlockSize write WriteBlockSize;
    property Rows: Byte read FRows write WriteRows;
    property Columns: Byte read FColumns write WriteColumns;
    property DrawGrid: Boolean read FDrawGrid write WriteDrawGrid;
    property Pieces: TStringList read ReadPieces write WritePieces;

    property OnLineFull: TOnLineFull read FOnLineFull write FOnLineFull;
    property OnGameEnd: TOnGameEnd read FOnGameEnd write FOnGameEnd;
  end;

procedure Register;

implementation

const
  DEF_BLOCKSIZE = 16;
  DEF_COLUMNS   = 9;
  DEF_ROWS      = 20;
  DEF_COLOR     = clBlack;
  DEF_GRIDCOLOR = clWhite;

  DIV_CHAR      = ';';

  FIELD_EMPTY   = 0;


procedure ClearMatrix(var M: TMatrix);
var I, J: Byte;
begin
  for I := 1 to MaxMatrixSize do
    for J := 1 to MaxMatrixSize do
      M[I, J] := FALSE;
end;

function FullMatrixString(M: TMatrix): String;
var i, j: Integer;
  S: String;
begin
  S := '';
  for J := 1 to MaxMatrixSize do
    begin
      for I := 1 to MaxMatrixSize do
        S := S + IntToStr(Ord(M[I, J]));

      S := S + #13;
    end;

  Result := S;
end;

function RotateMatrix(M: TMatrix; Size, Dir: Byte; Around: TPoint): TMatrix;
var I, J, nI, nJ: Integer;
  Src: Boolean;
begin
  ClearMatrix(Result);

  for I := 1 to Size do
    for J := 1 to Size do
      begin
        Src := M[I, J];

        if Dir = 1 then
          begin
            nI := Size - J + 1;
            nJ := I;
          end else
        if Dir = 0 then
          begin
            nI := J;
            nJ := Size - I + 1;
          end else
          begin
            nI := 1;
            nJ := 1;
          end;

        Result[nI, nJ] := Src;
      end;
end;

function FillMatrix(M: String; Size: Byte): TMatrix;
var CharIndex, TotSize, i, j: Byte;
  Ch: Char;
begin
  TotSize := Size * Size;
  while Length(M) < TotSize do
    M := M + '0';

  ClearMatrix(Result);

  for J := 1 to Size do
    for I := 1 to Size do
      begin
        CharIndex := (J - 1) * Size + I;
        Ch := M[CharIndex];
        Result[I, J] := Ch = '1';
      end;

end;

function GetParam(S, Divider: String; Param: Word): String;
var Count, P: Word;
  Prm: String;
begin
  S := S + Divider;
  Result := '';

  P := Pos(Divider, S);
  for Count := 1 to Param do
    begin
      Prm := Copy(S, 1, P - 1);
      Delete(S, 1, P);
      P := Pos(Divider, S);
    end;

  Result := Prm;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TKozyTetris]);
end;

constructor TKozyTetris.Create(AOwner: TComponent);
begin
  Inherited Create(AOwner);

  FPieces := TStringList.Create;
  with FPieces do
    begin
      Add('Stick;clWhite;4;0100010001000100');
      Add('Block;clRed;2;1111');
      Add('Z;clLime;3;110011000');
      Add('S;clAqua;3;011110000');
      Add('T;clPurple;3;010111000');
      Add('L;clBlue;3;010010011');
      Add('Inverse L;clYellow;3;010010110');
    end;

  FPlaying := FALSE;

  FColor := DEF_COLOR;
  FGridColor := DEF_GRIDCOLOR;

  {Write these two properties directly}
  FColumns := DEF_COLUMNS;
  FRows := DEF_ROWS;
  {and this one with the WriteBlockSize method, this way we initialize
   the Width and Height}
  BlockSize := DEF_BLOCKSIZE;
end;

destructor TKozyTetris.Destroy;
begin
  FPieces.Destroy;

  inherited Destroy;
end;

procedure TKozyTetris.WriteBlockSize(NewBlockSize: Byte);
begin
  if Playing then
    Exit;

  FBlockSize := NewBlockSize;

  Width := FColumns * FBlockSize + 1;
  Height := FRows * FBlockSize + 1;
end;

procedure TKozyTetris.WriteRows(NewRows: Byte);
begin
  if Playing then
    Exit;

  if NewRows < 2 then
    NewRows := 2;
  if NewRows > MaxPlayRows then
    NewRows := MaxPlayRows;

  FRows := NewRows;
  Height := FRows * FBlockSize + 1;
end;

procedure TKozyTetris.WriteColumns(NewColumns: Byte);
begin
  if Playing then
    Exit;

  if NewColumns < 2 then
    NewColumns := 2;
  if NewColumns > MaxPlayColumns then
    NewColumns := MaxPlayColumns;

  FColumns := NewColumns;
  Width := FColumns * FBlockSize + 1;
end;

procedure TKozyTetris.WriteColor(NewColor: TColor);
begin
  FColor := NewColor;
  Paint;
end;

procedure TKozyTetris.WriteGridColor(NewGridColor: TColor);
begin
  FGridColor := NewGridColor;
  Paint;
end;

procedure TKozyTetris.WriteDrawGrid(NewDrawGrid: Boolean);
begin
  FDrawGrid := NewDrawGrid;

  Paint;
end;

function TKozyTetris.ReadPieces: TStringList;
begin
  Result := FPieces;
end;

procedure TKozyTetris.WritePieces(NewPieces: TStringList);
begin
  if Playing then
    Exit;

  FPieces.Assign(NewPieces);
end;


procedure TKozyTetris.Paint;
var I, J: Integer;
begin
  inherited Paint;

  with Canvas do
    begin
      Brush.Color := FColor;
      //FillRect(Rect(0, 0, Width, Height));

      Pen.Width := 1;
      for I := 1 to FColumns do
        for J := 1 to FRows do
          begin
            if Field[I, J] <> FIELD_EMPTY then
              Brush.Color := TPiece(FPieceList[Field[I, J] - 1]).PieceColor else
              Brush.Color := FColor;

            FillRect( Rect(
              (I - 1) * FBlockSize + 1, (J-1) * FBlockSize + 1,
               I * FBlockSize, J * FBlockSize));
          end;

        Pen.Width := 1;
        for i := 0 to FColumns do
          begin
            MoveTo(i * FBlockSize, 0);
            LineTo(i * FBlockSize, Height + 1);
          end;

        for j := 0 to FRows do
          begin
            MoveTo(0, j * FBlockSize);
            LineTo(Width + 1, j * FBlockSize);
          end;

        for i := 1 to FColumns - 1 do
          for j := 1 to FRows - 1 do
            begin
              if FDrawGrid then
                Pixels[i * FBlockSize, j * FBlockSize] := FGridColor;
            end;

      if Playing then
        DrawCurrentPiece(pX, pY);
    end;
end;

procedure TKozyTetris.DrawPieceTo(Canvas: TCanvas; X, Y: Integer; Piece: TPiece; Erase: Boolean);
var R: TRect;
  I, J: Integer;
  P: TPiece;
begin
  if not Assigned(Piece) then
    Exit;

  P := Piece;

  for I := 1 to P.PieceSize do
    for J := 1 to P.PieceSize do
      if P.Matrix[I, J] then
        begin
          R := Rect(
            (X + I - 2) * FBlockSize + 1,
            (Y + J - 2) * FBlockSize + 1,
            (X + I - 1) * FBlockSize,
            (Y + J - 1) * FBlockSize);

          with Canvas do
            begin
              if Erase then
                Brush.Color := FColor else
                Brush.Color := P.PieceColor;

              FillRect(R);
            end;
        end;
end;

procedure TKozyTetris.DrawPiece(X, Y: Integer; Piece: TPiece);
begin
  DrawPieceTo(Canvas, X, Y, Piece, FALSE);
end;

procedure TKozyTetris.DrawCurrentPiece(X, Y: Integer);
begin
  DrawPieceTo(Canvas, X, Y, FCurrentPiece, FALSE);
end;

procedure TKozyTetris.ErasePiece(X, Y: Integer; Piece: TPiece);
begin
  DrawPieceTo(Canvas, X, Y, Piece, TRUE);
end;

procedure TKozyTetris.EraseCurrentPiece(X, Y: Integer);
begin
  DrawPieceTo(Canvas, X, Y, FCurrentPiece, TRUE);
end;

procedure TKozyTetris.ClearPieceList;
var All, i: Integer;
  P: TPiece;
begin
  All := FPieceList.Count;
  if All = 0 then
    Exit;

  for I := All - 1 downto 0 do
    begin
      P := TPiece(FPieceList[I]);
      P.Destroy;
      FPieceList.Delete(I);
    end;

  FPieceList.Destroy;
end;

function TKozyTetris.InitializePieces: Boolean;
var All, i: Integer;
  Mat, RotS, S: String;
  Piece: TPiece;
begin
  Result := FALSE;
  FCanPlay := FALSE;

  try
    if Assigned(FPieceList) then
      ClearPieceList;

    FPieceList := TList.Create;
  except
    Exit;
  end;

  FCanPlay := TRUE;

  All := FPieces.Count;
  if All = 0 then
    begin
      FCanPlay := FALSE;
      raise Exception.Create('No pieces to initialize.');
    end;

  for I := 0 to All - 1 do
    begin
      S := FPieces[I];
      if S = '' then
        Continue;

      Piece := TPiece.Create;
      with Piece do
        begin
          PieceName := GetParam(S, DIV_CHAR, 1);

          try
            PieceColor := StringToColor(GetParam(S, DIV_CHAR, 2));
          except
            PieceColor := clWhite;
          end;
          
          while PieceColor = FColor do
            PieceColor := Random(clWhite);

          try
            PieceSize := StrToInt(GetParam(S, DIV_CHAR, 3));
          except
            PieceSize := 4;
          end;

          if PieceSize < 1 then
            PieceSize := 1;
            
          if PieceSize > MaxMatrixSize then
            PieceSize := MaxMatrixSize;

          Mat := GetParam(S, DIV_CHAR, 4);
          Matrix := FillMatrix(Mat, PieceSize);

          RotS := GetParam(S, DIV_CHAR, 5);
          if RotS = '' then
            begin
              Rot := Point(0, 0);
            end else
          if RotS <> '' then
            begin
              try
                Rot := Point(
                  StrToInt(GetParam(RotS, ',', 1)),
                  StrToInt(GetParam(RotS, ',', 2)));
              except
                Rot := Point(0, 0);
              end;

              if (Rot.X < 0) or (Rot.X > MaxMatrixSize) or
                 (Rot.Y < 0) or (Rot.Y > MaxMatrixSize) then
                Rot := Point(0, 0);
            end;
        end;

      FPieceList.Add(Piece);
    end;

  Result := TRUE;
end;

procedure TKozyTetris.RotatePiece(Clockwise: Boolean);
begin
  if not Assigned(FCurrentPiece) then
    Exit;

  if not CanRotate(Clockwise) then
    Exit;

  EraseCurrentPiece(pX, pY);

  with FCurrentPiece do
    Matrix := RotateMatrix(Matrix, PieceSize, Ord(Clockwise), Point(0, 0));

  DrawCurrentPiece(pX, pY);
end;

function TKozyTetris.CanRotate(Clockwise: Boolean): Boolean;
var I, J, K, L: Integer;
  Stuck: Boolean;
  P: TPiece;
begin
  Result := FALSE;
  if (not Assigned(FCurrentPiece)) or (NOT Playing) then
    Exit;

  Stuck := FALSE;
  P := TPiece.Create;
  P.Assign(FCurrentPiece);
  with P do
    Matrix := RotateMatrix(Matrix, PieceSize, Ord(Clockwise), Point(0, 0));

  for I := 1 to P.PieceSize do
    for J := 1 to P.PieceSize do
      if P.Matrix[I, J] then
        begin
          K := pX + I - 1;
          L := pY + J - 1;

          if (K < 1) or (K > FColumns) or (L > FRows) or
             ((Field[K, L] <> FIELD_EMPTY) and (L > 0)) then
            begin
              Stuck := TRUE;
              Break;
            end;
        end;

  Result := NOT Stuck;
  P.Free;
end;

procedure TKozyTetris.StartNewGame;
begin
  Randomize;

  FPlaying := FALSE;
  ClearPlayField;
  Paint;

  if not InitializePieces then
    begin
      raise Exception.Create('Piece initialization failure; could not start a new game.');
      Exit;
    end;

  SetNewPiece;
  FPlaying := TRUE;
end;

function TKozyTetris.PieceCount: Integer;
begin
  Result := 0;
  if not Assigned(FPieceList) then
    Exit;

  Result := FPieceList.Count;
end;

procedure TKozyTetris.SetCurrentPieceIndex(Index: Integer);
begin
  if (Index < 0) or (Index >= PieceCount) then
    Exit;

  if not Assigned(FCurrentPiece) then
    FCurrentPiece := TPiece.Create;

  FCurrentPiece.Assign(TPiece(FPieceList[Index]));
  FCurrentPieceIndex := Index;
end;

procedure TKozyTetris.ClearPlayField;
var I, J: Integer;
begin
  for I := 1 to MaxPlayColumns do
    for J := 1 to MaxPlayRows do
      Field[I, J] := FIELD_EMPTY;
end;

function TKozyTetris.CanMove(X, Y: Integer): Boolean;
var I, J, K, L: Integer;
  Stuck: Boolean;
begin
  Result := FALSE;
  if (not Assigned(FCurrentPiece)) or (NOT Playing) then
    Exit;

  Stuck := FALSE;

  for I := 1 to FCurrentPiece.PieceSize do
    for J := 1 to FCurrentPiece.PieceSize do
      if FCurrentPiece.Matrix[I, J] then
        begin
          K := pX + I - 1 + X;
          L := pY + J - 1 + Y;

          if (K < 1) or (K > FColumns) or (L > FRows) or
             ((Field[K, L] <> FIELD_EMPTY) and (L > 0)) then
            begin
              Stuck := TRUE;
              Break;
            end;
        end;

  Result := NOT Stuck;
end;

function TKozyTetris.MoveLeft: Integer;
begin
  if CanMove(-1, 0) then
    begin
      EraseCurrentPiece(pX, pY);
      Inc(pX, -1);
      DrawCurrentPiece(pX, pY);
      Result := 1;
    end else
    Result := 0;
end;

function TKozyTetris.MoveRight: Integer;
begin
  if CanMove(1, 0) then
    begin
      EraseCurrentPiece(pX, pY);
      Inc(pX, 1);
      DrawCurrentPiece(pX, pY);
      Result := 1;
    end else
    Result := 0;
end;

function TKozyTetris.MoveDown: Integer;
begin
  if CanMove(0, 1) then
    begin
      EraseCurrentPiece(pX, pY);
      Inc(pY, 1);
      DrawCurrentPiece(pX, pY);
      Result := 1;
    end else
    begin
      if pY <= 0 then
        begin
          Result := -1;

          if Assigned(FOnGameEnd) and Playing then
            FOnGameEnd(Self);

          FPlaying := FALSE;
        end else
        Result := 0;
        
      MapPieceToMatrix;
      SetNewPiece;
      Paint;
    end;
end;

function TKozyTetris.DropDown: Integer;
var mY: Integer;
begin
  mY := 1;
  while CanMove(0, mY) do
    Inc(mY);

  EraseCurrentPiece(pX, pY);
  Inc(pY, mY - 2);
  DrawCurrentPiece(pX, pY);
  Result := 1;

  MoveDown;
end;

function TKozyTetris.MapPieceToMatrix: Boolean;
var I, J: Integer;
  P: TPiece;
begin
  Result := FALSE;
  if not Assigned(FCurrentPiece) then
    Exit;

  Result := TRUE;

  try
    P := FCurrentPiece;
    for I := 1 to P.PieceSize do
      for J := 1 to P.PieceSize do
        if P.Matrix[I, J] then
          Field[pX + I - 1, pY + J - 1] := FCurrentPieceIndex + 1;

    CheckFullLines;
    Paint;
  except
    Result := FALSE;
  end;
end;

function TKozyTetris.CountEmptyLines: Integer;
var I, J, All: Integer;
  Empty: Boolean;
begin
  All := 0;

  for J := 1 to FRows do
    begin
      Empty := TRUE;

      for I := 1 to FColumns do
        if Field[I, J] <> FIELD_EMPTY then
          Empty := FALSE;

      if Empty then
        Inc(All) else
        Break;
    end;

  FEmptyLines := All;
  Result := All;
end;

function TKozyTetris.SetNewPiece: integer;
var Ws, We, Hs, He, PieceWidth, PieceHeight, I, J: Integer;
begin
  SetCurrentPieceIndex(Random(PieceCount));

  We := MaxMatrixSize;
  He := MaxMatrixSize;
  
  Hs := 9;
  Ws := 9;

  for J := 1 to FCurrentPiece.PieceSize do
    for I := 1 to FCurrentPiece.PieceSize do
      begin
        if FCurrentPiece.Matrix[I, J] then
          begin
            if J < Hs then
              Hs := J else
              He := J;

            if I < Ws then
              Ws := I else
              We := I;
          end;
      end;

  PieceHeight := He - Hs + 1;
  PieceWidth := We - Ws + 1;
  
  pX := (FColumns div 2) - (PieceWidth div 2);
  pY := - PieceHeight + 2;
  Result := CurrentPiece;
end;

procedure TKozyTetris.CheckFullLines;
var Lines, X, Y, I, J: Integer;
  Full: Boolean;
begin
  Lines := 0;
  Y := FRows;
  repeat
    Full := TRUE;
    for X := 1 to FColumns do
      if Field[X, Y] = FIELD_EMPTY then
        begin
          Full := FALSE;
          Break;
        end;

    if Full then
      begin
        for J := Y downto 2 do
          for I := 1 to FColumns do
            Field[I, J] := Field[I, J - 1];

        for I := 1 to FColumns do
          Field[I, 1] := FIELD_EMPTY;

        Inc(Lines);
      end else
      Dec(Y);
  until Y <= 0;

  if (Lines > 0) and Assigned(FOnLineFull) then
    FOnLineFull(Self, Lines);

end;



procedure TPiece.Assign(Source: TPiece);
begin
  PieceColor := Source.PieceColor;
  PieceSize := Source.PieceSize;
  PieceName := Source.PieceName;
  Matrix := Source.Matrix;
  Rot := Source.Rot;
end;


end.

