unit OInspector;
{ This is an component to browse the published properties of any Object
  thats inherited from TPersistent: TObjectInspector.
  It is similar to the Delphi-Objectinspector.
  You are able to change the object properties at Runtime.

  The component based on the idea from "Adam Cutchin" who designed the
  Tf_Inspector (Last Version on 10-22-96)

  I associated other components like the InplaceEdit from "MIKSoft, Inc."
  and this compacted unit is the result of all.

  Therefor, I have some problems to free the memory after the used of
  TObjectInspector. To solve the problem temporally, You have to call the
  FreeMemory method in the EventHandler "OnDestroy" of Your Form.
  I used Memcheck to detected memory leaks

  TObjectInspector is an Freeware component with Source and not copyrighted.
  You can do what ever You wnat with this component.
  USE IT AT YOUR OWN RISK

  Version: 0.15; 8 Sept. 1999

  However, I hope this component can helps You.
  Please notify me if You know how to solve the FreeMemory problem
  or other unknown bugs.

  eMail: jmk@mb2.tu-chemnitz.de

  To do:
     - Property-Editor for TStrings and TFont
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ComCtrls, Buttons, TypInfo, InEdit2;

type
  TShowEvent = Procedure(Sender : Tobject;  var CanShow : Boolean) of Object;

  PropEdType = (peProp, peSetItem);

  PPropEditData = ^TPropEditData;
  TPropEditData = record
    Kind: PropEdType;
    Depth: Integer;
    FObject: TObject;
    PInfo: PPropInfo;
    LastStr: String[75];
    case PropEdType of
      peSetItem: (Value: LongInt; Name: ShortString);
      peProp: (PropData: PPropInfo; ChildShowing: Boolean);
  end;

  TEdControl = class(TInplaceEdit)
  protected
    procedure DoExit; override;
  end;

  TBtnControl = class(TSpeedButton)
  protected
    procedure Click; override;
  end;

  TLBoxControl = class(TCustomListBox)
  protected
    BtnDown: Boolean;
    procedure DoExit; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  end;

  TObjectInspector = class(TCustomListBox)
  private
    HalfWay: Integer;
    SplitDrag : Boolean;

    EdControl: TEdControl;
    BtnControl: TBtnControl;
    LbxControl: TLBoxControl;

    EditRect : TRect;
    DownArrow : TBitmap;
    ComboButtonBox : TRect;
    EditHit : Real;
    CBDown : Boolean;

    FAutoCalcSep: Boolean; // autocalc Separator line
    procedure SetAutoCalcSep(Value: Boolean);
  protected
    { Set and Get functions for the properies }
    procedure DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton;  Shift: TShiftState; X, Y: Integer); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure DblClick; override;
    procedure CreateParams(var Params: TCreateParams); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function IsA(PInfo: PTypeInfo; S: string): Boolean;
    function SeeIfChanged(X : Integer) : Boolean;
    procedure CheckProperties(AObject: TObject);
    procedure FreeMemory;
  published
    { pusblish the TListBox properties }
    property AutoCalcSep: Boolean read FAutoCalcSep write SetAutoCalcSep;
    property Align;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ItemHeight;
    property Items;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Sorted;
    property Style default lbOwnerDrawFixed;
    property TabOrder;
    property Visible;
    property Width;
    { pusblish the TListBox events }
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

var
  Editors : Array[1..1000] of Record
    PropType : TClass;
    EditIt   : TNotifyEvent;
  End;

  NumEditors : Integer;
  FontSender : Tobject;

  function IsRegistered(C : TObject) : Boolean;
  procedure Edit(O : TObject);

procedure Register;

implementation

procedure Edit(O: TObject);
var
  x : integer;
Begin
  For x := 1 to NumEditors do
  if O is Editors[X].PropType  then
  begin
    Editors[X].EditIt(O);
    Break;
  end;
end;

Function IsRegistered(C : TObject): Boolean;
var x : integer;
Begin
  for x := 1 to NumEditors do
  if C is Editors[X].PropType then
  begin
    IsRegistered := True;
    Exit;
  end;
  IsRegistered := False;
end;

{ TEdControl }
procedure TEdControl.DoExit;
Var
  PPED : PPropEditData;             { property to what was entered into the }
  X : Integer;                                                  { edit box. }
  NewInt : Integer;
  NewReal : Real;
begin
  If Modified then
  Begin
    with (Parent as TObjectInspector) do
      PPED := Pointer(Items.Objects[ItemIndex]);
    If PPED^.Kind = peProp then
    Case PPED^.PropData^.PropType^.Kind of
      tkInteger:
        Begin
          Val(Text, NewInt, X);
          if x = 0 then
            SetOrdProp(PPED^.FObject, PPED^.PropData, NewInt)
          Else Messagebeep(0);
        End;
      tkFloat:
        Begin
          Val(Text, NewReal, X);
          if x = 0 then
            SetFloatProp(PPED^.FObject, PPED^.PropData, NewReal)
          Else Messagebeep(0);
        End;
      tkEnumeration:
        Begin
          x := GetEnumValue(PPED^.PropData^.PropType^, Text);
          if x  <> -1 then
            SetOrdProp(PPED^.FObject, PPED^.PropData, X)
          else Messagebeep(0);
        End;
      tkString, tkLString, tkWString:
        SetStrProp(PPED^.FObject, PPED^.PropData, Text);
    End Else
      MessageBeep(0);
  end;
End;

{ TBtnControl }
procedure TBtnControl.Click;
begin
  (Parent as TObjectInspector).DblClick;
end;

{ TLBoxControl }
procedure TLBoxControl.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  PPED: PPropEditData;
begin
  if not btnDown then
  Begin
    SendMessage(Handle, wm_LButtonDown, mk_Lbutton, y shl 16 + x);
    sendMessage(Handle, wm_LButtonUp, mk_Lbutton, y shl 16 + x);
  End;
  PPED := Pointer((Parent as TObjectInspector).
    Items.Objects[(Parent as TObjectInspector).ItemIndex]);
  If PPED^.Kind = peProp then
  Begin
    SetOrdProp(PPED^.FObject, PPED^.PropData,
      GetEnumValue(PPED^.PropData^.PropType^, Items[ItemIndex]));
  (Parent as TObjectInspector).SetFocus;
  End;
end;

procedure TLBoxControl.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  If (X >= ClientRect.left) and (X <= ClientRect.Right) and
     (y >= ClientRect.Top) and (Y <= ClientRect.Bottom) then
  Begin
    BtnDown := True;
  End Else Begin
  End;
end;

procedure TLBoxControl.DoExit;
begin
  Visible := False;
  (Parent as TObjectInspector).CBDown := False;
  (Parent as TObjectInspector).SetFocus;
end;

{ TObjectInspector }
procedure TObjectInspector.CreateParams;
begin
  inherited CreateParams(Params); { call the inherited first }
//  with Params do Style := Style or WS_CAPTION;// or WS_SIZEBOX; { add a style flag }
  with Params do ExStyle := ExStyle or WS_EX_TOOLWINDOW;
end;

constructor TObjectInspector.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FAutoCalcSep := True;
  
  ItemHeight := 18;
  Color := clBtnFace;
  Style := lbOwnerDrawFixed;
  { edit control }
  EdControl := TEdControl.Create(Self);
  EdControl.BorderStyle := bsNone;
  EdControl.Parent := Self;
  EdControl.Visible := False;
  { Button control }
  BtnControl := TBtnControl.Create(Self);
  BtnControl.Parent := Self;
  BtnControl.Visible := False;
  { Listbox control }
  LbxControl := TLBoxControl.Create(Self);
  LbxControl.Ctl3D := False;
  LbxControl.ItemHeight := 18;
  LbxControl.Parent := Self;
  LbxControl.Visible := False;
  { Down arrow }
  DownArrow := TBitmap.Create;
  DownArrow.Width := 7;
  DownArrow.Height := 4;
  DownArrow.canvas.Brush.Color := clBtnFace;
  DownArrow.canvas.Pen.Color := clBtnface;
  DownArrow.Canvas.Rectangle(0,0,7,4);
  DownArrow.Canvas.Pen.Width := 0;
  DownArrow.Canvas.pen.Color := clBlack;
  DownArrow.Canvas.MoveTo(0, 0);
  DownArrow.Canvas.LineTo(7, 0);
  DownArrow.Canvas.MoveTo(1, 1);
  DownArrow.Canvas.LineTo(6, 1);
  DownArrow.Canvas.MoveTo(2, 2);
  DownArrow.Canvas.LineTo(5, 2);
  DownArrow.Canvas.Pixels[3, 3] := clBlack;
end;

destructor TObjectInspector.Destroy;
begin
  DownArrow.Free;
  BtnControl.Free;
  EdControl.Free;
  LbxControl.Free;
  inherited Destroy;
end;

procedure TObjectInspector.FreeMemory;
var
  i: Integer;
begin
  for i := 0 to Items.Count - 1 do
    FreeMem(Pointer(Items.Objects[i]), SizeOf(TPropEditData));
end;

procedure TObjectInspector.SetAutoCalcSep(Value: Boolean);
begin
  if FAutoCalcSep <> Value then
    FAutoCalcSep := Value;
end;

procedure TObjectInspector.CheckProperties(AObject: TObject);
var
  PL: PPropList;
  Size : Integer;
  X : Integer;
  PPED : PPropEditData;
begin
  if EdControl.Focused then
    SetFocus;
  BtnControl.Visible := false;
  FreeMemory;
  Clear;
  Try
    Size := GetTypeData(AObject.Classinfo)^.PropCount;
    Visible := True;
  Except
    Visible := False;
    Exit;
  End;
  GetMem(PL, Size*SizeOf(Pointer));
  try
    GetPropInfos(AObject.ClassInfo, PL);
    HalfWay := 0;
    if not FAutoCalcSep then
      HalfWay := Width div 2;
    for X := 0 to Size-1 do
    if pl[x] <> nil then
    if pl[x]^.PropType^.Kind <> tkMethod then
    Begin
      GetMem(PPED, SizeOf(TPropEditData));
      PPED^.FObject := AObject;
      PPED^.PInfo := @pl[x]^.PropType;
      PPED^.Kind := peProp;
      PPED^.ChildShowing := False;
      PPED^.Depth := 0;
      PPED^.PropData := pl[x];
      if FAutoCalcSep then
      if Canvas.TextWidth(PPED^.PInfo.Name)+10 > HalfWay then
        HalfWay := Canvas.TextWidth(PPED^.PInfo.Name)+10;
      SeeIfChanged(Items.AddObject('', TObject(PPED)));
    end;
  finally
    FreeMem(PL, Size*SizeOf(Pointer));
  end;
end;

{ Like 'is,' but works with a string. }
function TObjectInspector.IsA(PInfo: PTypeInfo; S: string): Boolean;
var
  PTData : PTypeData;
  PTInfo : PTypeInfo;
  Found : Boolean;
Begin
  Found := False;
  S := UpperCase(S);
  PTInfo := PInfo;
  while (not Found) and (UpperCase(PTInfo^.Name) <> 'TOBJECT') do
  begin
    Found := UpperCase(PTInfo^.Name) = S;
    PTData := GetTypeData(PTInfo);
    PTInfo := PTData^.ParentInfo^;
  End;
  IsA := Found;
End;

function TObjectInspector.SeeIfChanged(X : Integer): Boolean;
var
  Obj1: TObject;
  Obj2: TObject;
  PInfo: PPropInfo;
  RString: string;
  PPED: PPropEditData;
  V: Integer;
Begin
  PPED := Pointer(Items.Objects[X]);
  Obj1 := PPED^.FObject;
  try
    if PPED^.Kind = peProp then
    begin
      PInfo := PPED^.PropData;
      If PPED^.Kind = peProp then
      begin
        If Obj1 <> Nil then
        case PInfo^.Proptype^.Kind of
          tkInteger, tkChar: Rstring := IntToStr(GetOrdProp(Obj1, PInfo));
          tkSet: RString := '(Set...)';
          tkEnumeration :
            Begin
              V := GetOrdProp(Obj1, PInfo);
              RString := GetEnumName(PInfo^.PropType^, V);
            End;
          tkFloat: RString := FloatToStrf(GetFloatProp(Obj1, PInfo), ffnumber, 5, 10);
          tkString, tkLString: RString := GetStrProp(Obj1, PInfo);
          tkClass:
            Begin
              Obj2 := TObject(GetOrdProp(Obj1, PInfo));
              If Assigned(Obj2) and (Obj2 is TComponent) then
                RString := TComponent(Obj2).Name
              else
                RString := GetTypeData(PInfo^.Proptype^)^.ClassType.ClassName;
              if not Assigned(Obj2) then
                RString := RString + ' (Nil)';
            End;
        End;
      End;

      If PPED^.Kind = peSetItem then
      if (GetOrdProp(PPropEditData(PPED^.FObject)^.FObject, PPED^.PInfo) and
        (1 shl PPED^.Value)) >= 1 then
        RString := 'True'
      Else
        RString := 'False';
    End;
  except
    RString := '**Exception**';
  end;
  Result :=  PPED^.LastStr <> RString;
//  PPED^.LastStr := RString;
End;

procedure TObjectInspector.DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);
var
  PInfo: PPropInfo;
  Object1: TObject;
  EdString, RString: string;
  V: Integer;
  PEditData: PPropEditData;
  S: String;
  Object2: TObject;
  SRect: TRect;
  DRect: TRect;
begin
  if Assigned(OnDrawItem) then
    OnDrawItem(Self, Index, Rect, State)
  else
  begin
//    HalfWay := Width div 2;
    PEditData := Pointer(Items.Objects[Index]);
    HideCaret(Handle);
    Object1 := PEditData^.FObject;
    S := '';
    for V := 0 to PEditData^.Depth do
      S := S + '  ';
    if PEditData^.Kind = peProp then
    begin
      PInfo := PEditData^.PropData;
      If PInfo^.PropType^.kind in [tkSet, tkClass] then
        If (PInfo^.PropType^.kind = tkClass) then
        begin
          S := Copy(S, 1, Length(S) - 2);
          if IsA(PInfo^.PropType^, 'TObject') then
            S := S + '  '
          else
          if PEditData^.ChildShowing then
            S := S + '- '
          else
            s := s + '+';
        end else
        begin
          S := Copy(S, 1, Length(S) - 2);
          if PEditData^.ChildShowing then
            S := S + '- '
          else
            S := S + '+';
      end;
    end;
    { draw the item }
    with Canvas do
    begin
      Brush.Color := Self.Color;  {Won't use Windows Selected color.}
      SRect := Rect;
      If odSelected in state then
        SRect.Right := HalfWay;
      FillRect(SRect);
      if (odSelected in State) then
        Pen.Color := clBtnHighLight
      else {for 3D look}
        Pen.Color := clBtnShadow;
      Pen.Mode := pmCopy;
      Pen.Width := 1;
      Pen.Style := psSolid;
      MoveTo(rect.Left, Rect.Bottom-1);
      LineTo(rect.Right, Rect.Bottom-1);
      if (odSelected in State) then
      begin
        Pen.Color := clBtnHighLight;
        MoveTo(HalfWay + 2, Rect.Top+1);
        LineTo(HalfWay + 2, Rect.Bottom);

        Pen.Color := clBtnShadow;
        MoveTo(HalfWay+1, Rect.Top+2);
        LineTo(HalfWay+1, Rect.Bottom);

        MoveTo(Rect.Right, Rect.Top);
        LineTo(Rect.Left, Rect.Top);
        LineTo(Rect.Left, Rect.Bottom);

        Pen.Color := clBlack;
        MoveTo(Rect.Right - 1, Rect.Top + 1);
        LineTo(Rect.Left + 1, Rect.Top + 1);
        LineTo(Rect.Left + 1, Rect.Bottom - 1);
      end else
      begin
        Pen.Color := clBtnShadow;
        MoveTo(HalfWay, Rect.Top); //Top+1
        LineTo(HalfWay, Rect.Bottom);

        Pen.Color := clBtnHighLight;
        MoveTo(HalfWay + 1, Rect.Top); // Top+1
        LineTo(HalfWay + 1, Rect.Bottom);
      end;

      SRect.Left := Rect.Left;
      SRect.Right := HalfWay;
      SRect.Top := Rect.Top+1;
      SRect.Bottom := Rect.Bottom-1;
      Font.Color := clBlack;

      If (odSelected in State) then
      Begin
        Inc(SRect.Left);
        Inc(SRect.Top);
        Inc(SRect.Right);
      End;

      setBkMode(Canvas.Handle, Transparent);

      Case PEditData^.Kind of
        peProp: TextRect(SRect, SRect.Left, SRect.Top, S + PInfo^.Name);
        peSetItem: TextRect(SRect, SRect.Left, SRect.Top, S + PEditData^.Name);
      End;

      Pen.Color := clBtnFace;
      Font.Color := clNavy;

      try
        if PEditData^.Kind = peProp then
        begin
          if Object1 <> nil then
          case PInfo^.Proptype^.Kind of
            tkInteger, tkChar: Rstring := IntToStr(GetOrdProp(Object1, PInfo));
            tkSet: RString := '(Set...)';
            tkEnumeration :
              begin
                V := GetOrdProp(Object1, PInfo);
                RString := GetEnumName(PInfo^.PropType^, V);
              end;
            tkFloat: RString := FloatToStrf(GetFloatProp(Object1, PInfo), ffNumber, 5, 10);
            tkString, tkLString: RString := GetStrProp(Object1, PInfo);
            tkClass:
              begin
                Object2 := TObject(GetOrdProp(Object1, PInfo));
                If Assigned(Object2) and (Object2 is TComponent) then
                  RString := TComponent(Object2).Name
                else
                  RString := Format('(%s)',
                    [GetTypeData(PInfo^.Proptype^)^.ClassType.ClassName]);
                if not Assigned(Object2) then
                  RString := RString + ' (Nil)';
              end;
          end;
        end;

        if PEditData^.Kind = peSetItem then
        if (GetOrdProp(PPropEditData(PEditData^.FObject)^.FObject, PEditData^.PInfo) and
          (1 shl PEditData^.Value)) >= 1 then
          RString := 'True'
        Else
          RString := 'False';
      except
        RString := '** Exception **';
      End;

      PEditData^.LastStr := RString;

  Canvas.Font.Color := clNavy; // Text color on the right side (property value-str)
//  Font.Style := Font.Style + [fsBold];

  SRect := ItemRect(ItemIndex);
  if odSelected in state then
    if PEditData^.Kind = PeProp then      { Enumeration - Simulate List box. }
      if PInfo^.PropType^.kind = tkEnumeration then
      begin
        ComboButtonBox := SRect;
        SRect.Right := SRect.Right - ABS(SRect.Bottom - SRect.Top);
        Inc(ComboButtonBox.Top, 2);
        ComboButtonBox.Left := SRect.Right + 1;
        DrawButtonFace(Canvas, ComboButtonBox, 3, bsNew, false, false, false);

        DRect.Left := ComboButtonBox.Left +
          ((ComboButtonBox.Right - ComboButtonBox.Left) DIV 2) - (7 DIV 2);
        DRect.Right := DRect.Left + DownArrow.Width;
        DRect.Top := ComboButtonBox.Top +
          ((ComboButtonBox.Bottom - ComboButtonBox.Top) DIV 2) - (4 DIV 2);
        DRect.Bottom := DRect.Top + DownArrow.Height;

        Canvas.CopyRect(DRect, DownArrow.Canvas, DownArrow.Canvas.ClipRect);

        BtnControl.visible := False;
      End;

      if (not EdControl.Focused) and (odSelected in State) then
      With Canvas do
      Begin
        EdControl.Visible := False;
        EditRect := ItemRect(ItemIndex);
        EditRect.Left := HalfWay + 2;
        Brush.Color := clWhite;
        Inc(EditRect.Top, 2);
        if PEditData^.Kind = peProp then
        if PInfo^.PropType^.kind = tkEnumeration then
          EditRect.Right := EditRect.Right - ABS(EditRect.Bottom - EditRect.Top);
        if (PEditData.Kind = peProp) and (PInfo^.PropType^.Kind = tkClass) then
        if Assigned(Object2) then
        begin
          if IsRegistered(Object2) then
          begin
            BtnControl.Width := EditRect.Bottom - EditRect.Top;
            EditRect.Right := EditRect.Right - ABS(EditRect.Bottom - EditRect.Top);
            BtnControl.Left := Editrect.Right + 1;
            BtnControl.Top := EditRect.Top;
            BtnControl.Height := EditRect.Bottom - EditRect.Top;
            BtnControl.Visible := true;
          End else
            BtnControl.Visible := False;
        End else
          BtnControl.Visible := False;
        Fillrect(EditRect);
        setBkColor(Canvas.Handle, GetSysColor(color_HighLight));
        setBkMode(Canvas.Handle, Opaque);
        Font.Color := clHighlightText;
        Canvas.Brush.Color := clHighlight;
        EdString := RString;
        If (PEditData.Kind = peProp) and
          (PInfo^.PropType^.kind in [tkSet, tkEnumeration, tkClass]) then
          EdControl.ReadOnly := True
        else
          EdControl.ReadOnly := False;

        while (TextWidth(EdString) > (EditRect.Right - EditRect.Left)) and
          (EdString <> '')  do
          Delete(EdString, 1, 1);
        TextOut(EditRect.Left + 2, EditRect.Top+1, edString);
        DestroyCaret;
        CreateCaret(Handle, 0, 1, (EditRect.Bottom - Editrect.Top) - 2);
        SetCaretPos(EditRect.Left + TextWidth(EdString), EditRect.Top + 1);
        ShowCaret(Handle);
        EdControl.Text := RString;
      end else
      begin
        if (odSelected in state) and CBDown then
        Begin
          Canvas.Brush.Color := clWhite;
          Canvas.FillRect(EditRect);
          Font.Color := clWindowText;
          If EdControl.Focused then
            RString := EdControl.Text;
          Canvas.TextOut(HalfWay + 2, Rect.Top+2, RString);
        end else;
        Canvas.TextOut(HalfWay + 3, Rect.Top+1, RString);
      End;
      if (odFocused in State) and (odSelected in State) then
        Canvas.DrawFocusRect(Rect);

      If BtnControl.Visible then
        BtnControl.Refresh;
    end;
  end;
end;

procedure TObjectInspector.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  P: TPoint;
  PPED: PPropEditData;
  PI: PPropInfo;
  PTD: PTypeData;
begin
  If Abs(x - Halfway) <= 3 then
    SplitDrag := True
  else
  If (X >= EditRect.left) and (X <= EditRect.Right) and
     (y >= EditRect.Top) and (Y <= EditRect.Bottom) then
  Begin
    P.X := x;
    p.y := y;
    Cursor := crDefault;
    EdControl.left := EditRect.left;
    EdControl.Top := EditRect.Top;
    EdControl.Width := EditRect.Right - EditRect.Left;
    EdControl.Height := EditRect.Bottom - EditRect.Top;
    Edithit := Time;
    EdControl.Visible := True;
    MapWindowPoints(Handle, EdControl.Handle, P, 1);
    PostMessage(EdControl.Handle, WM_LButtonDown, MK_LButton, P.Y shl 16 + P.X);
  End Else
  If (X >= ComboButtonBox.left) and (X <= ComboButtonBox.Right) and
    (y >= ComboButtonBox.Top) and (Y <= ComboButtonBox.Bottom) and (not CBDown) then
  Begin
    PPED := Pointer(Items.Objects[ItemIndex]);
    PI := PPED^.PropData;
    PTD := GetTypeData(PI^.PropType^);
    LbxControl.Visible := True;
    EdControl.Visible := False;
    BtnControl.Visible := False;
    CBDown := True;
    LbxControl.Items.Clear;

    LbxControl.Top := EditRect.Bottom;
    LbxControl.Left := EditRect.Left;
    LbxControl.Width := ComboButtonBox.Right-EditRect.Left;
    for x := PTD^.MinValue to PTD^.maxValue do
      LbxControl.Items.Add(GetEnumname(PI^.PropType^, X));
    if LbxControl.Items.Count > 6 then
      LbxControl.Height := 4 + LbxControl.ItemHeight *  6
    else
      LbxControl.Height := 4 + LbxControl.ItemHeight * LbxControl.Items.Count;
   LbxControl.SetFocus;
  End;
end;

procedure TObjectInspector.MouseMove(Shift: TShiftState; X, Y: Integer);
{ handles dragging the split in the two panes. }
Var
  TempR : TRect;
begin
  If SplitDrag then
  Begin
    If X > ClientWidth - 20 then
      X := ClientWidth - 20;
    If X < 20 then
      X := 20;
    If Halfway < x then
    Begin
      TempR.Left := Halfway;
      TempR.Right := ClientWidth;
      TempR.Top := 0;
      TempR.Bottom := ClientHeight;
    End Else
    Begin
      TempR.Left := X;
      TempR.Right := ClientWidth;
      TempR.Top := 0;
      TempR.Bottom := ClientHeight;
    End;
    HalfWay := x;
    InvalidateRect(Handle, @tempR, false);
    InvalidateRect(Handle, @EditRect, False);
    PostMessage(Handle, WM_Paint, 0, 0);
    Exit;
  End;
  If Abs(x - HalfWay) <= 3 then
    Cursor := crHSplit
  Else
  If (X >= EditRect.left) and (X <= EditRect.Right) and
     (y >= EditRect.Top) and (Y <= EditRect.Bottom) then
    Cursor := crIBeam else Cursor := crDefault;
End;

procedure TObjectInspector.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  SplitDrag := False;
end;

procedure TObjectInspector.DblClick;
var
  PD : PTypeData;
  PD1 : PTypeData;
  NewPPED : PPRopEditData;
  PS : ^ShortString;
  PPED : PPropEditData;
  Size: Integer;
  X : Integer;
  PL: PPropList;
  o : TObject;
  JP : Pointer;
  P : Tpoint;

  procedure HideChildren(I : Integer);
  var
    PE : PPropEditData;
    X : integer;
    JP : Pointer;
    PE2 : PPropEditData;
  begin
    PE := Pointer(Items.Objects[I]);
    if PE^.Kind = peSetItem then Exit;
    JP := nil;
    case PE^.PropData^.PropType^.Kind of
      tkClass : JP := Pointer(getOrdProp(PE^.FObject, PE^.PInfo));
      tkSet   : JP := PE;
    end;
    if JP = nil then exit;
    O := TObject(JP);
    if PE^.ChildShowing then
    begin
      PE^.childShowing := False;
      X := 0;
      while X < Items.Count do
      begin
        if (PPropEditData(Items.Objects[x])^.FObject = O) then
        begin
          PE2 := Pointer(Items.Objects[X]);
          If PE2^.ChildShowing then
            HideChildren(X);
          FreeMem(Pointer(Items.Objects[x]), Sizeof(TPropEditData));
          Items.Delete(X);
          Dec(X);
        end;
        Inc(X);
      end;
    end;
  end;

begin
  GetCursorPos(P);
  MapWindowPoints(0, Handle, P, 1);
  PPED := Pointer(Items.Objects[ItemIndex]);
  { Handle clicks left of the break. }
  { Basically show or hide children properties. }
  If P.X < halfway then
  Begin
    if PPED^.Kind = peProp then
    Begin
      JP := nil;
      Case PPED^.PropData^.PropType^.Kind of
        tkClass : JP := Pointer(getOrdProp(PPED^.FObject, PPED^.PInfo));
        tkSet   : JP := PPED;
      End;
      If JP = Nil then exit;
      O := Tobject(JP);
      If PPED^.ChildShowing then
      Begin
        HideChildren(ItemIndex);
        Exit;
      End;
      Case PPED^.PropData^.PropType^.Kind of
        tkClass:
          Begin
            Size := GetTypeData(PPED^.PropData^.PropType^)^.PropCount;
            GetMem(PL, Size*SizeOf(Pointer));
            try
              GetPropInfos(PPED^.PropData^.PropType^, PL);
              PPED^.ChildShowing := True;
              for X := 0 to Size-1 do
              if PL[x]^.PropType^.Kind <> tkMethod then
              Begin
                GetMem(NewPPED, SizeOf(TPropEditData));
                NewPPED^.FObject := TObject(GetOrdProp(PPED^.FObject, PPED^.PInfo));
                NewPPED^.Kind := peProp;
                NewPPED^.ChildShowing := False;
                NewPPED^.Depth := PPED^.Depth + 1;
                NewPPED^.PropData := PL[x];
                NewPPED^.Kind := peProp;
                Items.InsertObject(ItemIndex+1 + X,'', TObject(newPPED));
              End;
            finally
              FreeMem(PL, Size * SizeOf(Pointer));
            end;  
          End;
        tkSet:
          Begin
            PD := GetTypeData(PPED^.PropData^.PropType^);
            if PD^.CompType^.Kind = tkEnumeration then
            Begin
              PD1 := GetTypeData(PD^.CompType^);
              PPED^.ChildShowing := True;
              PS := @PD1^.NameList;
              For x := PD1^.MinValue to PD1^.MaxValue do
              Begin
                GetMem(NewPPED, SizeOf(TPropEditData));
                NewPPED^.Value := X;
                NewPPEd^.Name := PS^;
                Inc(Integer(PS), Length(PS^) + 1);
                NewPPED^.Depth := PPED^.Depth + 1;
                NewPPED^.FObject := Tobject(PPED);
                NewPPED^.PInfo := PPED^.PropData;
                NewPPED^.Kind := peSetItem;
                Items.InsertObject(ItemIndex+1+(X - PD1^.MinValue),'',TObject(newPPED));
              End;
            End;
          End;
      End;
    End;          { Now handle clicks right of the divide. }
  End Else
  Begin
    If PPED^.kind = peProp then
    Case PPED^.PropData^.PropType^.Kind of
      tkClass:
        Begin
          JP := Pointer(getOrdProp(PPED^.FObject, PPED^.PInfo));
          If JP = Nil then exit;
          O := Tobject(JP);
          If IsRegistered(o) then
            Edit(o);
        End;
      tkEnumeration:
        Begin
          PPED := Pointer(Items.Objects[ItemIndex]);
          PD := GetTypeData(PPED^.PropData.PropType^);
          x := getOrdProp(PPED^.FObject, PPED^.PropData);
          Inc(x);
          if x > PD^.maxvalue then x := pd^.minvalue;
          SetOrdProp(PPED^.FObject, PPED^.Propdata, X);
        End;
      End Else
      Begin  { A set option. }
        If PPED^.Kind = peSetItem then
          SetOrdProp(PPropEditData(PPED^.FObject)^.FObject, PPED^.PInfo,
          GetOrdProp(PPropEditData(PPED^.FObject)^.FObject, PPED^.PInfo) XOR (1 shl PPED^.Value));
      End;
  End;
  SetFocus;
  Refresh;
end;

procedure TObjectInspector.KeyDown(var Key: Word; Shift: TShiftState);
begin
  If (Key = VK_RETURN) or (Key = VK_SPACE) then
  begin
    Click;
    Key := 0;
    Exit;
  end;
  if not (Key in [VK_UP, VK_DOWN, VK_PRIOR, VK_NEXT]) then
  begin
    EdControl.left := EditRect.left;
    EdControl.Top := EditRect.Top;
    EdControl.Width := EditRect.Right - EditRect.Left;
    EdControl.Height := EditRect.Bottom - EditRect.Top;
    EdControl.Visible := True;
    EdControl.SetFocus;
    PostMessage(EdControl.Handle, WM_KeyDown, key, 0);
    Key := 0;
  End;
end;

procedure Register;
begin
  RegisterComponents('DocJohn', [TObjectInspector]);
end;

Begin
  NumEditors := 0;
end.
