{-------------------------------------------------------------------------------}
{                                                                               }
{ Freeware. Please mail me if you use this, or have any suggestions.            }
{ hughdoar@hotmail.com                                                          }
{                                                                               }
{-------------------------------------------------------------------------------}

{
 hwd 23.1.99 v 1.0
     26.1.99   1.1 added event linkage
      3.2.99   1.2 to use hwdStringGrid instead
               1.3 fix for col widths
     14.9.99   1.4 fixed filter
}

unit hwdListView;

interface

uses Grids, hwdTreeView, hwdStringGrid,
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls;

type
  { Clears and fills a ListView with data from the Grid according to the }
  { position of the current item in the Tree. The Tree Grid should be    }
  { the same length as the List Grid. Note the first line of the         }
  { grid is taken to be headings, and is ignored. Also note the event    }
  { redirection from FTree.OnChange.                                     }
  ThwdListView = class(TListView)
  private
    FGrid :ThwdStringGrid;
    FTree :ThwdTreeview;
    procedure FilterFillFromGrid;
  protected
    procedure SetTree(ahwdTreeView :ThwdTreeView);
    procedure SetGrid(aStringGrid :ThwdStringGrid);
    procedure TreeChange(Sender: TObject; Node: TTreeNode);
  public
    constructor Create(aComponent :TComponent); override;
    procedure Clear;
  published
    property Tree :ThwdTreeView read FTree write SetTree;
    property Grid :ThwdStringGrid read FGrid write SetGrid;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('External', [ThwdListView]);
end;

constructor ThwdListView.Create(aComponent :TComponent);
begin
  inherited Create(aComponent);
  ViewStyle := vsReport;
end;

procedure ThwdListView.Clear;
begin
  Items.Clear;
  Columns.Clear;
end;

procedure ThwdListView.SetTree(ahwdTreeView :ThwdTreeView);
begin
  FTree := ahwdTreeView;
  FilterFillFromGrid;
  if (FTree <> nil) then FTree.OnChangeForList := TreeChange;
end;

procedure ThwdListView.SetGrid(aStringGrid :ThwdStringGrid);
var i :Integer;
begin
  Clear;
  FGrid := aStringGrid;
  if (FGrid <> nil) then
  begin
    for i := 0 to FGrid.ColCount - 1 do
      with Columns.Add do
      begin
        Caption := FGrid.Cells[i, 0];
        Width := FGrid.ColWidths[i];
      end;
    FilterFillFromGrid;
  end;
end;

procedure ThwdListView.TreeChange(Sender: TObject; Node: TTreeNode);
begin
  { could pass this as the selected node... }
  FilterFillFromGrid;
end;

procedure ThwdListView.FilterFillFromGrid;
var slFilters :TStringList;
    aNode :TTreeNode;
    i, j :Integer;
    filterOK :Boolean;
    aListItem :TListItem;
begin
  if (FGrid = nil) then Exit;
  Items.Clear;
  { if there is a tree, get our filter list, of length = tree level }
  slFilters := TStringList.Create;
  if (FTree <> nil) and (FTree.Grid <> nil) then
  begin
    if (FGrid.RowCount <> FTree.Grid.RowCount) then
      raise Exception.Create('TreeView Grid and ListView Grid are different lengths.');
    aNode := FTree.Selected;
    while (aNode <> nil) do
    begin
      slFilters.Insert(0, aNode.Text);
      aNode := aNode.Parent;
    end;
  end;
  { the below ignores the first line of the grid (taken to be headings), and }
  { the first line of the filters, which is the root node, 'all'.            }
  for j := 1 to FGrid.RowCount - 1 do
  begin
    { decide whether to include this line... }
    filterOK := True;
    if (FTree <> nil) and (FTree.Grid <> nil) then
      if slFilters.Count > 1 then
        for i := 1 to slFilters.Count - 1 do
          if (FTree.Grid.Cells[i - 1, j] <> slFilters[i]) then
          begin
            filterOK := False;
            Break;
          end;
    if filterOK then
    begin
      aListItem := Items.Add;
      aListItem.Caption := FGrid.Cells[0, j];
      for i := 1 to FGrid.ColCount - 1 do
        aListItem.SubItems.Add(FGrid.Cells[i, j]);
    end;
  end;
  slFilters.Free;
end;




end.
