{-------------------------------------------------------------------------------}
{                                                                               }
{ Freeware. Please mail me if you use this, or have any suggestions.            }
{ hughdoar@hotmail.com                                                          }
{                                                                               }
{-------------------------------------------------------------------------------}

{
 hwd  1.2.99 v 1.0
      2.2.99   1.1 added dataset, patched InsertColumn for empty grid
      3.2.99   1.2 added row functions and drag drop capabilities
      14.9.99  1.3 fixed clear datasource bug
               1.4 changed DataSource to Public
}

unit hwdStringGrid;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, DB;

type
  ThwdStringGrid = class(TStringGrid)
  private
    FPreviousSortList :TStringList;
    FAscending :Boolean;
    FDataSource :TDataSource;
    FDragCol, FDragRow :Integer;
  protected
    procedure SetDataSource(aDataSource :TDataSource);
    procedure DragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    constructor Create(aComponent :TComponent); override;
    destructor Destroy; override;
    procedure DragDrop(Source: TObject; X, Y: Integer); override;
    procedure DeleteRow(Index :Integer); override;
    procedure Clear;
    procedure MoveCol(FromIndex, ToIndex :Integer);
    procedure MoveRow(FromIndex, ToIndex :Integer);
    procedure SortCol(Index :Integer; Ascending :Boolean);
    procedure SortRow(Index :Integer; Ascending :Boolean);
    procedure DeleteCol(Index :Integer);
    procedure ExchangeCol(IndexA, IndexB :Integer);
    procedure ExchangeRow(IndexA, IndexB :Integer);
    procedure InsertCol(aCol :TStringList; Index, aColWidth :Integer);
    procedure InsertRow(aRow :TStringList; Index, aRowHeight :Integer);
    { FDragCol and FDragRow indicate the state of drag for a host }
    {        -1           -1             nothing                  }
    {         x           -1             col x being dropped in   }
    {        -1            x             row x being dropped in   }
    { and are set exclusively in MouseDown                        }
    property DragCol :Integer read FDragCol;
    property DragRow :Integer read FDragRow;
    { the below cannot be set at design time, as although it will work in }
    { the designer, at create the dataset may not have been created.      }
    property DataSource :TDataSource read FDataSource write SetDataSource;
  published
  end;

const
  BADINT     :Integer = -99999999;

procedure Register;


implementation


procedure Register;
begin
  RegisterComponents('External', [ThwdStringGrid]);
end;

{ ThwdStringGrid }

constructor ThwdStringGrid.Create(aComponent :TComponent);
begin
  inherited Create(aComponent);
  FPreviousSortList := TStringList.Create;
  FDragCol := -1;
  FDragRow := -1;
  FAscending := True;

  FixedCols := 0;
  FixedRows := 0;
  ColCount := 0;
  RowCount := 0;
  Options := [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine,
              goColSizing, goRowSizing,
              goColMoving, goRowMoving,
              goDrawFocusSelected, goThumbTracking];
  DefaultColWidth := 32;
end;

destructor ThwdStringGrid.Destroy;
begin
  FPreviousSortList.Free;
  inherited Destroy;
end;

procedure ThwdStringGrid.Clear;
begin
  RowCount := 1;
  ColCount := 1;
  Cells[0, 0] := '';
end;

procedure ThwdStringGrid.DeleteCol(Index :Integer);
var i :Integer;
begin
  if (Index < 0) or (Index >= ColCount) then Exit;
  if (Index < ColCount - 1) then
    for i := Index to ColCount - 2 do
    begin
      Cols[i].Assign(Cols[i + 1]);
      ColWidths[i] := ColWidths[i + 1];
    end;
  if (Index = 0) and (ColCount = 1) then Clear else ColCount := ColCount - 1;
end;

procedure ThwdStringGrid.DeleteRow(Index :Integer);
var i :Integer;
begin
  if (Index < 0) or (Index >= RowCount) then Exit;
  if (Index < RowCount - 1) then
    for i := Index to RowCount - 2 do
    begin
      Rows[i].Assign(Rows[i + 1]);
      RowHeights[i] := RowHeights[i + 1];
    end;
  if (Index = 0) and (RowCount = 1) then Clear else RowCount := RowCount - 1;
end;

procedure ThwdStringGrid.ExchangeCol(IndexA, IndexB :Integer);
var i :Integer;
begin
  if (IndexA < 0) or (IndexA > ColCount - 1) or
     (IndexB < 0) or (IndexB > ColCount - 1) or (IndexA = IndexB) then Exit;
  for i := 0 to RowCount - 1 do Rows[i].Exchange(IndexA, IndexB);
end;

procedure ThwdStringGrid.ExchangeRow(IndexA, IndexB :Integer);
var i :Integer;
begin
  if (IndexA < 0) or (IndexA > RowCount - 1) or
     (IndexB < 0) or (IndexB > RowCount - 1) or (IndexA = IndexB) then Exit;
  for i := 0 to ColCount - 1 do Cols[i].Exchange(IndexA, IndexB);
end;

procedure ThwdStringGrid.InsertCol(aCol :TStringList; Index, aColWidth :Integer);
{ no idea why the w is needed. it's a fix. }
var i, w :Integer;
begin
  w := ColWidths[ColCount - 1];
  if aColWidth < 0 then aColWidth := DefaultColWidth;
  if (ColCount = 1) and (RowCount = 1) then RowCount := aCol.Count
    else ColCount := ColCount + 1;
  if (Index >= ColCount) then Index := ColCount - 1;
  if (Index < 0) then Index := 0;
  if (Index < ColCount - 1) and (ColCount <> 1) then
    for i := ColCount - 1 downto Index + 1 do
    begin
      Cols[i].Assign(Cols[i - 1]);
      ColWidths[i] := ColWidths[i - 1];
    end;
  if ColCount = 1 then Index := 0;
  Cols[Index].Assign(TStrings(aCol));
  ColWidths[Index] := aColWidth;
  ColWidths[ColCount - 2] := w;
end;

procedure ThwdStringGrid.InsertRow(aRow :TStringList; Index, aRowHeight :Integer);
{ no idea why the h is needed. it's a fix. }
var i, h :Integer;
begin
  h := RowHeights[RowCount - 1];
  if aRowHeight < 0 then aRowHeight := DefaultRowHeight;
  if (RowCount = 1) and (ColCount = 1) then ColCount := aRow.Count
    else RowCount := RowCount + 1;
  if (Index >= RowCount) then Index := RowCount - 1;
  if (Index < 0) then Index := 0;
  if (Index < RowCount - 1) and (RowCount <> 1) then
    for i := RowCount - 1 downto Index + 1 do
    begin
      Rows[i].Assign(Rows[i - 1]);
      RowHeights[i] := RowHeights[i - 1];
    end;
  if RowCount = 1 then Index := 0;
  Rows[Index].Assign(TStrings(aRow));
  RowHeights[Index] := aRowHeight;
  RowHeights[RowCount - 2] := h;
end;

procedure ThwdStringGrid.SortCol(Index :Integer; Ascending :Boolean);
{ though this does surprisingly well, it really needs better auto-sensing,
  detection of field type from a datasource, and it doesn't maintain order
  on previously sorted columns...yet.
  it also sorts the fixed rows :( }
var i :Integer;
    TS :TStringList;
  function FixForInteger(s :String) :String;
  begin
    s := '0000000000' + s; //10
    result := copy(s, length(s) - 9, 10);
  end;
begin
  if (Index < 0) or (Index >= ColCount) then Exit;
  TS := TStringList.Create;
  TS.Assign(Cols[Index]);
  for i := 0 to TS.Count - 1 do
  begin
    TS.Objects[i] := TStringList.Create;
    TStringList(TS.Objects[i]).Assign(Rows[i]);
  end;
  for i := 0 to TS.Count - 1 do
    if StrToIntDef(TS[i], BADINT) <> BADINT then TS[i] := FixForInteger(TS[i]);
  TS.Sorted := True;
  for i := 0 to TS.Count - 1 do
  begin
    if Ascending then Rows[i].Assign(TStringList(TS.Objects[i]))
    else Rows[TS.Count - 1 - i].Assign(TStringList(TS.Objects[i]));
    TStringList(TS.Objects[i]).Free;
  end;
  TS.Free;
end;

procedure ThwdStringGrid.SortRow(Index :Integer; Ascending :Boolean);
{ - above routine will change... }
var i :Integer;
    TS :TStringList;
  function FixForInteger(s :String) :String;
  begin
    s := '0000000000' + s; //10
    result := copy(s, length(s) - 9, 10);
  end;
begin
  if (Index < 0) or (Index >= RowCount) then Exit;
  TS := TStringList.Create;
  TS.Assign(Rows[Index]);
  for i := 0 to TS.Count - 1 do
  begin
    TS.Objects[i] := TStringList.Create;
    TStringList(TS.Objects[i]).Assign(Cols[i]);
  end;
  for i := 0 to TS.Count - 1 do
    if StrToIntDef(TS[i], BADINT) <> BADINT then TS[i] := FixForInteger(TS[i]);
  TS.Sorted := True;
  for i := 0 to TS.Count - 1 do
  begin
    if Ascending then Cols[i].Assign(TStringList(TS.Objects[i]))
    else Cols[TS.Count - 1 - i].Assign(TStringList(TS.Objects[i]));
    TStringList(TS.Objects[i]).Free;
  end;
  TS.Free;
end;

procedure ThwdStringGrid.MoveCol(FromIndex, ToIndex :Integer);
var i :Integer;
    TS :TStringList;
begin
  if (ToIndex < 0) or (ToIndex >= ColCount) then ToIndex := ColCount - 1;
  if (FromIndex = ToIndex) or (FromIndex < 0) or (FromIndex >= ColCount) then Exit;
  TS := TStringList.Create;
  TS.Assign(Cols[FromIndex]);
  if (FromIndex < ToIndex) then
    for i := FromIndex to ToIndex - 1 do Cols[i].Assign(Cols[i + 1])
  else
    for i := FromIndex downto ToIndex + 1 do Cols[i].Assign(Cols[i - 1]);
  Cols[ToIndex].Assign(TStrings(TS));
  TS.Free;
end;

procedure ThwdStringGrid.MoveRow(FromIndex, ToIndex :Integer);
var i :Integer;
    TS :TStringList;
begin
  if (ToIndex < 0) or (ToIndex >= RowCount) then ToIndex := RowCount - 1;
  if (FromIndex = ToIndex) or (FromIndex < 0) or (FromIndex >= RowCount) then Exit;
  TS := TStringList.Create;
  TS.Assign(ROws[FromIndex]);
  if (FromIndex < ToIndex) then
    for i := FromIndex to ToIndex - 1 do Rows[i].Assign(Rows[i + 1])
  else
    for i := FromIndex downto ToIndex + 1 do Rows[i].Assign(Rows[i - 1]);
  Rows[ToIndex].Assign(TStrings(TS));
  TS.Free;
end;

procedure ThwdStringGrid.SetDataSource(aDataSource :TDataSource);
{ also use to refresh }
var i, j :Integer;
begin
  if (aDataSource = nil) or (aDataSource.DataSet = nil) then
  begin
    Clear;
    fDataSource := nil;
  end
    else
  begin
    try
      with aDataSource.DataSet do
      begin
        if not Active then Open;
        RowCount := RecordCount + 1;
        ColCount := FieldCount;
        FixedCols := 0;
        FixedRows := 1;
        for i := 0 to FieldCount - 1 do Cells[i, 0] := Fields[i].DisplayName;
        First;
        j := 1;
        while not Eof do
        begin
          for i := 0 to FieldCount - 1 do
            Cells[i, j] := Fields[i].AsString;
          j := j + 1;
          Next;
        end;
        fDataSource := aDataSource;
      end;
    except
      Clear;
    end;
  end;
end;

procedure ThwdStringGrid.DragDrop(Source: TObject; X, Y: Integer);
var DroppedOnIdx, DroppedIdx :Integer;
    aCol, aRow :TStringList;
    aColWidth, aRowHeight :Integer;
begin
  if (ThwdStringGrid(Source).DragRow = -1) then
  begin {it's a col}
    DroppedOnIdx := DragCol;
    if DroppedOnIdx = -1 then DroppedOnIdx := ColCount;
    DroppedIdx := ThwdStringGrid(Source).DragCol;
    aCol := TStringList(ThwdStringGrid(Source).Cols[DroppedIdx]);
    aColWidth := ThwdStringGrid(Source).ColWidths[DroppedIdx];
    InsertCol(aCol, DroppedOnIdx, aColWidth);
    if ColCount = 1 then FixedRows := 1;
    ThwdStringGrid(Source).DeleteCol(DroppedIdx);
  end
    else
  begin {it's a row}
    DroppedOnIdx := DragRow;
    if DroppedOnIdx = -1 then DroppedOnIdx := RowCount;
    DroppedIdx := ThwdStringGrid(Source).DragRow;
    aRow := TStringList(ThwdStringGrid(Source).Rows[DroppedIdx]);
    aRowHeight := ThwdStringGrid(Source).RowHeights[DroppedIdx];
    InsertRow(aRow, DroppedOnIdx, aRowHeight);
    if RowCount = 1 then FixedCols := 1;
    ThwdStringGrid(Source).DeleteRow(DroppedIdx);
  end;
  ThwdStringGrid(Source).EndDrag(True);
  ThwdStringGrid(Source).Refresh;
  inherited DragDrop(Source, X, Y);
end;

procedure ThwdStringGrid.DragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
begin
  Accept := False;
  inherited DragOver(Source, X, Y, State, Accept);
  if not((Source is ThwdStringGrid) and (Source <> Self)) then Exit;
    if (ThwdStringGrid(Source).DragCol > -1) or
       (ThwdStringGrid(Source).DragRow > -1) then
    begin
      { the fields set below are the row/col dropped *on* in this receiver }
      MouseToCell(X, Y, FDragCol, FDragRow);
      Accept := True;
    end;
end;

procedure ThwdStringGrid.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
end;

procedure ThwdStringGrid.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  MouseToCell(X, Y, FDragCol, FDragRow);
  if (FDragCol = -1) or (FDragRow = -1) then Exit;
  if (FDragCol > FixedCols - 1) and (FDragRow > FixedRows - 1) then
  begin
    inherited MouseDown(Button, Shift, X, Y);
  end;

  if (ssShift in Shift) then
  begin  { sort on column }
    SortCol(FDragCol, FAscending);
    FAscending := not FAscending;
    Exit;
  end;

  if (ssCtrl in Shift) then
  begin { external drag }
    { exit if trying to drag x,y from grid with <x fixed col and <y fixed row }
    if (FDragRow < FixedRows) and (FDragCol < FixedCols) and
       ((FixedCols >= 1) and (FixedRows >= 1)) then Exit;
    { indicate to host which type is being dropped }
    if (FixedCols = 0) then FDragRow := -1;
    if (FixedRows = 0) then FDragCol := -1;
    if (FixedCols >= 1) and (FixedRows >= 1) and (FDragCol < FixedCols) then FDragCol := -1;
    if (FixedCols >= 1) and (FixedRows >= 1) and (FDragRow < FixedRows) then FDragRow := -1;
    { now DragR/C is -1 if [C/R] is being moved, and DragC/R is the index }
    BeginDrag(False);
    Exit;
  end;

  if (Screen.Cursor = crHSplit) or (Screen.Cursor = crVSplit) then
  begin
    inherited MouseDown(Button, Shift, X, Y);
    Exit;
  end;

  { otherwise, we must be trying to move the column internally... }
  inherited MouseDown(Button, Shift, X, Y);
end;

end.
