{-------------------------------------------------------------------------------}
{                                                                               }
{ Freeware. Please mail me if you use this, or have any suggestions.            }
{ hughdoar@hotmail.com                                                          }
{                                                                               }
{-------------------------------------------------------------------------------}

{
 hwd 23.1.99 v 1.0
     26.1.99   1.1 added event linkage
     27.1.99   1.2 root child bug fixed (inefficient)
      3.2.99   1.3 to use hwdStringGrid instead
      4.2.99   1.4 added 'proper' root node
}

unit hwdTreeView;

interface

uses hwdStringGrid,
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, Grids;

type
  { Associates a hwdStringGrid with the TreeView. When this grid is set,      }
  { the TreeView is updated to reflect it. Note the first line of the         }
  { grid is taken to be headings, and is ignored.                             }
  { Use an entire column of duplicates to provide a 'root' node.              }
  { NB - if you free the grid that Grid points to before setting Grid to nil, }
  { then Grid will point to nowhere. Disconnect the link b4 freeing the grid. }

  ThwdTreeView = class(TTreeView)
  private
    fGrid :ThwdStringGrid;
    { the below is set by the ListView }
    fOnChangeForList :TTVChangedEvent;
    fNoSpaces :Boolean;
    function GetChildNode(aNode :TTreeNode; s :String) :TTreeNode;
    function CreateChildNode(aNode :TTreeNode; s :String) :TTreeNode;
    function CreateRootNode(s :String) :TTreeNode;
    procedure FillFromGrid;
  protected
    procedure SetGrid(aStringGrid :ThwdStringGrid);
    { override this to give us time to tell the ListView that we've changed }
    procedure Change(Node: TTreeNode); override;
    procedure DragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean); override;
  public
    constructor Create(aComponent :TComponent); override;
    procedure DragDrop(Source: TObject; X, Y: Integer); override;
    procedure Clear;
    property OnChangeForList :TTVChangedEvent read fOnChangeForList write fOnChangeForList;
  published
    { setting this fills the TreeView }
    property Grid :ThwdStringGrid read fGrid write SetGrid;
    { set this to ignore spaces in the grid. set the grid again. }
    property NoSpaces :Boolean read fNoSpaces write fNoSpaces;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('External', [ThwdTreeView]);
end;

{ ThwdTreeView }

constructor ThwdTreeView.Create(aComponent :TComponent);
begin
  inherited Create(aComponent);
  fGrid := nil;
  fNoSpaces := False;
end;

procedure ThwdTreeView.Clear;
begin
  Items.Clear;
end;

procedure ThwdTreeView.SetGrid(aStringGrid :ThwdStringGrid);
begin
  fGrid := aStringGrid;
  FillFromGrid;
end;

function ThwdTreeView.CreateChildNode(aNode :TTreeNode; s :String) :TTreeNode;
begin
  result := Items.AddChild(aNode, s);
end;

function ThwdTreeView.CreateRootNode(s :String) :TTreeNode;
begin
  result := Items.Add(nil, s);
end;

procedure ThwdTreeView.FillFromGrid;
{ this is a little slow! }
var i, j, idx :Integer;
    aNode, aParent, aRootNode :TTreeNode;
    sl :TStringList;
begin
  Clear;
  if fGrid = nil then Exit;
  aRootNode := CreateRootNode('All');
  { skip top row }
  for j := 1 to fGrid.RowCount - 1 do
  begin
    if (fNoSpaces) and (fGrid.Cells[0, j] = '') then Continue;
    aNode := aRootNode;
    for i := 0 to fGrid.ColCount - 1 do
    begin
      aParent := aNode;
      if (fNoSpaces) and (fGrid.Cells[i, j] = '') then Continue;
      aNode := GetChildNode(aNode, fGrid.Cells[i, j]);
      if aNode = nil then aNode := CreateChildNode(aParent, fGrid.Cells[i, j]);
    end
  end;
end;

function ThwdTreeView.GetChildNode(aNode :TTreeNode; s :String) :TTreeNode;
var i :Integer;
begin
  result := nil;
  for i := 0 to aNode.Count - 1 do
    if aNode.Item[i].Text = s then result := aNode.Item[i];
end;

procedure ThwdTreeView.Change(Node: TTreeNode);
begin
  if Assigned(fOnChangeForList) then fOnChangeForList(Self, Node);
  inherited Change(Node);
end;

procedure ThwdTreeView.DragDrop(Source: TObject; X, Y: Integer);
var aItem: TTreeNode;
    aAttachMode: TNodeAttachMode;
    aHitTests: THitTests;
begin
  aHitTests := GetHitTestInfoAt(X, Y);
  aItem := GetNodeAt(X, Y);
  if (htOnItem in aHitTests) or (htOnIcon in aHitTests) then
  begin
    aAttachMode := naAddChild;
    Selected.MoveTo(aItem, aAttachMode);
  end;
  inherited DragDrop(Source, X, Y);
end;

procedure ThwdTreeView.DragOver(Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
var aHitTests : THitTests;
begin
  inherited DragOver(Source, X, Y, State, Accept);
  { could check inherited accept here }
  Accept := False;
  if (Selected = nil) then Exit;
  if (Source <> Self) then Exit;
  aHitTests := GetHitTestInfoAt(X, Y);
  if (htOnItem in aHitTests) or (htOnIcon in aHitTests) then Accept := True;
end;

end.
