unit kbmMemTable;

interface

uses SysUtils,Classes,Db,DbCommon;

// TKbmMemTable v. 1.22
const COMPONENT_VERSION = '1.22';
// =========================================================================
// An inmemory temporary table.
// Can be used as a demonstration of how to create descendents of TDataSet,
// or as in my case, to allow a program to generate temporary data that can
// be used directly by all data aware controls.
//
// Copyright 1999 Kim Bo Madsen/Optical Services - Scandinavia
// All rights reserved.
//
// LICENSE AGREEMENT
// You are allowed to used this component in any project for free.
// You are NOT allowed to claim that you have created this component or to
// copy its code into your own component and claim that is was your idea.
// Im offering this for free for your convinience, and the ONLY thing I REQUIRE
// is to get an e-mail about what project this component (or derived versions)
// is used for. That will be my reward of offering this component for free to you!
//
// You dont need to state my name in your software, although it would be
// appreciated if you do.
//
// If you find bugs or alter the component (f.ex. see suggested enhancements
// further down), please DONT just send the corrected/new code out on the internet,
// but instead send it to me, so I can put it into the official version. You will
// be acredited if you do so.
//
//
// DISCLAIMER
// By using this component or parts theirof you are accepting the full
// responsibility of the use. You are understanding that the author cant be
// made responsible in any way for any problems occuring using this component.
// You also recognize the author as the creator of this component and agrees
// not to claim otherwize!
//
// Please forward corrected versions (source code ONLY!), comments,
// and emails saying you are using it for this or that project to:
//            kbm@optical.dk
//
// Suggestions for future enhancements:
//
//      - Support for SetKey, FindKey, GotoKey, FindNearest
//      - IDE designer for adding static data to the memtable.
//
// History:
//
//1.00:	The first release. Was created due to a need for a component like this.
//                                                                    (15. Jan. 99)
//1.01:	The first update. Release 1.00 contained some bugs related to the ordering
//	of records inserted and to bookmarks. Problems fixed.         (21. Jan. 99)

//1.02:	Fixed handling of NULL values. Added SaveToStream, SaveToFile,
//	LoadFromStream and LoadFromFile. SaveToStream and SaveToFile is controlled
//	by a flag telling if to save data, contents of calculated fields,
//	contents of lookupfields and contents of non visible fields.
//	Added an example application with Delphi 3 source code.       (26. Jan. 99)
//
//1.03: Claude Rieth from Computer Team sarl (clrieth@team.lu) came up with an
//      implementation of CommaText and made a validation check in _InternalInsert.
//      Because I allready have implemented the saveto.... functions, I decided
//      to implement Claude's idea using my own saveto.... functions. (27. Jan. 99)
//      I have decided to rename the component, because Claude let me know that
//      the RX library have a component with the same name as this.
//      Thus in the future the component will be named TkbmMemTable.
//      SaveToStream and LoadFromStream now set up date and decimal separator
//      temporary to make sure that the data saved can be loaded on another
//      installation with different date and decimal separator setups.
//      Added EmptyTable method to clear the contents of the memory table.
//
//1.04: Wagner ADP (wagner@cads-informatica.com.br) found a bug in the _internalinsert
//      procedure which he came up with a fix for.                     (4. Feb. 99)
//      Added support for the TDataset protected function findrecord.
//      Added support for CreateTable, DeleteTable.
//
//1.05: Charlie McKeegan from Task Software Limited (charlie@task.co.uk) found
//      a minor bug (and came up with a fix) in SetRecNo which mostly is
//      noticeable in a grid where the grid wont refresh when the record slider
//      is moved to the end of the recordset.                          (5. Feb. 99)
//      Changed SaveToStream to use the displayname of the field as the header
//      instead of the fieldname.
//
//1.06: Introduced a persistence switch and a reference to a file. If the
//      persistence switch is set, the file will be read aut. on table open,
//      and the contents of the table will be saved in the table on table close.
//
//1.07: Changed calculation of fieldofsets in InternalOpen to follow the fielddefs
//      instead of fields. It has importance when rearranging fields.
//      Because of this change the calculation of fieldoffsets into the buffer
//      has been changed too. These corrections was found to be needed to
//      support the new components tkbmPooledQuery and tkbmPooledStoredProc
//      (found in the package tkbmPooledConn)
//      which in turn handles pooled connections to a database. Very usefull
//      in f.ex a WWW application as they also makes the BDE function threadsafe,
//      and limits the concurrent connections to a database.
//
//1.08: Changed buffer algorithm in GetLine since the old one was faulty.
//      Problem was pointed out by: Markus Roessler@gmx.de
//
//1.09: Added LoadFromDataset, SaveToDataset, CreateTableAs,
//      BCD and BLOB support by me.
//      James Baile (James@orchiddata.demon.co.uk) pointed out a bug in GetRecNo
//      which could lead to an endless loop in cases with filtering. He also
//      provided code for sorting, which I have been rearranging a bit and
//      implemented.
//      Travis Diamond (tdiamond@airmail.net) pointed out a bug in GetWord where
//      null fields would be skipped.
//      Claudio Driussi (c.driussi@popmail.iol.it) send me version including
//      a sort routine, and Ive used some of he's ideas to implement a sorting
//      mechanism. Further he's code contained MoveRecords and MoveCurRec which
//      I decided to include in a modified form for Drag and Drop record
//      rearrangements.
//
//1.10: Support for Locate.                                             (17. May. 99)
//      Claudio Driussi (c.driussi@popmail.iol.it) came up with a fix for
//      GetRecNo. MoveRecord is now public.
//      Andrius Adamonis (andrius@prototechnika.lt) came up with fix for
//      call to validation routines in SetFieldData and support for
//      OldValue and NewValue in GetActiveRecordBuffer.
//
//1.11: Pascalis Bochoridis from SATO S.A Greece (pbohor@sato.gr)
//      Corrected bookmark behavior (substituted RecordNo from TRecInfo with unused Bookmark)
//      Corrected GetRecNo & SetRecNo  (Scrollbars now work ok at First and Last record)
//      Added LocateRecord, Locate, Lookup (I decided to use his Locate instead of my own).
//
//1.12: Added CloseBlob. Corrected destructor.                          (26. May. 99)
//      Corrected GetFieldData and SetFieldData. Could result in overwritten
//      memory elsewhere because of use of DataSize instead of Size.
//      Pascalis Bochoridis from SATO S.A Greece (phohor@sato.gr) send me a corrected
//      LocateRecord which solves problems regarding multiple keyfields.
//      Thomas Bogenrieder (tbogenrieder@wuerzburg.netsurf.de) have suggested a faster
//      LoadFromStream version which is much faster, but loads the full stream contents into
//      memory before processing. I fixed a few bugs in it, rearranged it and decided
//      to leave it up to you what you want to use.
//      I suggest that if you need speed for smaller streams, use his method, else use mine.
//      You can activate his method by uncommenting the define statement FAST_LOADFROMSTREAM
//      a few lines down.
//
//1.13  Corrected SetRecNo and MoveRecord.                               (31. May. 99)
//      By Pascalis Bochoridis from SATO S.A Greece (phohor@sato.gr)
//
//1.14  Respecting DefaultFields differently.                            (3. Jun. 99)
//      LoadFromDataset now includes a CopyStructure flag.
//      Supporting Master/Detail relations.
//      Now the Sort uses a new property for determining the fields to sort on,
//      which means its possible to sort on other fields than is used for
//      master/detail. Added SortOn(FieldNames:string) which is used for quick
//      addhoc sorting by just supplying the fieldnames as parameters to the
//      SortOn command.
//      Fixed memory leak in LoadFromDataset (forgot to free bookmark).
//      Added CopyFieldProperties parameter to LoadFromDataset and CreateTableAs which
//      if set to TRUE will copy f.ex. DisplayName, DisplayFormat, EditMask etc. from
//      the source.
//      Corrected OnValidate in SetFieldData. Was placed to late to have any impact.
//      Now checking for read only fields in SetFieldData.
//
//1.15  Totally rearranged LocateRecord to handle binary search when     (23. Jun. 99)
//      the data searched is sorted.
//      I was inspired by Paulo Conzon (paolo.conzon@smc.it) who send me a version with the Locate method
//      hacked for binary search. New methods PopulateRecord and PopulateField is used to create
//      a key record to search for.
//      Changed Sort and SortOn to accept sortoptions by a parameter instead of having a property.
//      The following sort options exists: mtsoCaseInsensitive and mtsoDescending.
//      mtsoPartialKey is used internally from LocateRecord where the TLocateOptions are remapped to
//      TkbmMemTableCompareOptions.
//
//1.16  Bug fixed in SaveToDataset. Close called where it shouldnt.      (19. Jul. 99)
//      Bug reported by Reinhard Kalinke (R_Kalinke@compuserve.com)
//      Fixed a few bugs + introduced Blob saving/loading in LoadFromStream/SaveToStream and thus
//      also LoadFromFile/SaveToFile. To save blob fields specify mtfSaveBlobs in save flags.
//      Full AutoInc implementation (max. 1 autoinc field/table).
//      These fixes and enhancements was contributed by Arsne von Wys (arsene@vonwyss.ch)
//      SetFieldData triggered the OnChange event to often. This is now fixed. Bug discovered by
//      Andrius Adamonis (andrius@prototechnika.lt).
//      Added mtfSkipRest save flag which if set will ONLY write out the fields specified by the rest of
//      the flags, while default operation is to put a marker to indicate a field skipped.
//      Usually mtfSkipRest shouldnt be specified if the stream should be reloaded by LoadFromStream later on.
//      But for generating Excel CSV files or other stuff which doesnt need to be reloaded,
//      mtfSkipRest can be valuable.
//      Greatly speeded up LoadFromStream (GetLine and GetWord).
//
//1.17  Supporting fieldtypes ftFixedChar,ftWideString.
//      Added runtime fieldtype checking for catching unsupported fields.
//      Raymond J. Schappe (rschappe@isthmus-ts.com) found a bug in CompareFields which he send a fix for.
//      Added a read only Version property.
//      The faster LoadFromStream added in 1.12 has now been removed due to the optimization of
//      the original LoadFromStream in 1.16. Tests shows no noticably performance difference anymore.
//      Inspired by Bruno Depero (bdepero@usa.net) which send me some methods for saving and
//      loading table definitions, I added mtfSaveDef to TkbmMemTableSaveFlag. If used the table
//      definition will be saved in a file. To copy another datasets definition, first do a CreateTableAs,
//      then SaveToFile/SaveToStream with mtfSaveDef.
//      Renamed TkbmSupportedFieldTypes to kbmSupportedFieldTypes and TkbmBlobTypes to kbmBlobTypes.
//      Generalized Delphi/BCB definitions.
//      Denis Tsyplakov (den@vrn.sterling.ru) suggested two new events: OnLoadField and OnLoadRecord
//      which have been implemented.
//      Added OnCompressSave and OnDecompressLoad for user defined (de)compression of SaveToStream,
//      LoadFromStream, SaveToFile and LoadFromFile.
//      Bruno Depero (bdepero@usa.net) inspired me to do this, since he send me a version including
//      Zip compression. But I like to generalize things and not to depend on any other specific
//      3. part library. Now its up to you which compression to use.
//      Added OnCompressBlobStream and OnDecompressBlobStream for (de)compression of inmemory blobs.
//      Added LZH compression to the Demo project by Bruno Depero (bdepero@usa.net).
//
//1.18  Changed SaveToStream and LoadFromStream to code special characters in string fields
//      (ftString,ftWideString,ftFixedChar,ftMemo,ftFmtMemo).
//      You may change this behaviour to the old way by setting kbmStringTypes to an empty set.
//      Fixed severe blob null bug. Blobs fields was sometimes reported IsNull=true even if
//      they had data in them.
//      Fixed a few minor bugs.
//
//1.19  Fixed a bug in CodedStringToString where SetLength was to long.             (10. Aug. 1999)
//      Bug reported by Del Piero (tomli@mail.tycoon.com.tw).
//      Fixed bug in LoadFromStream where DuringTableDef was wrongly initialized
//      to true. Should be false. Showed when a CSV file without definition was loaded.
//      Bug reported by Mr. Schmidt (ISAT.Schmidt@t-online.de)
//
//1.20  Marcelo Roberto Jimenez (mroberto@jaca.cetuc.puc-rio.br) reported a few bugs(23. Aug. 1999)
//      to do with empty dataset, which he also corrected (GetRecNo, FilterMasterDetail).
//      Furthermore he suggested and provided code for a Capacity property, which
//      can be used to preallocate room in the internal recordlist for a specific
//      minimum number of records.
//      Explicitly typecasted @ to PChar in several places to avoid problem about
//      people checking 'Typed @' in compile options.
//
//1.21  Corrected Locate on filtered recordssets.                                   (24. Aug. 1999)
//      Problem observed by Keith Blows (keithblo@woollyware.com)
//
//1.22  Corrected GetActiveRecord and added overridden SetBlockReadSize to be       (30. Aug. 1999)
//      compatible with D4/BCB4's TProvider functionality. The information and
//      code has been generously supplied by Jason Wharton (jwharton@ibobjects.com).
//      Paul Moorcroft (pmoor@netspace.net.au) added SaveToBinaryFile, LoadFromBinaryFile,
//      SaveToBinaryStream and LoadFromBinaryStream. They save/load the contents incl.
//      structure to/from the stream/file in a binary form which saves space and is
//      faster.
//      Added support for ftLargeInt (D4/D5/BCB4).
//=============================================================================

{$ifdef VER100}      // Delphi 3
{$define LEVEL3}
{$define DELPHI}
{$endif}

{$ifdef VER110}      // C++ Builder 3
{$define LEVEL3}
{$define BCB}
{$endif}

{$ifdef VER120}      // Delphi 4
{$define LEVEL4}
{$define DELPHI}
{$endif}

{$ifdef VER125}      // C++ Builder 4
{$define LEVEL4}
{$define BCB}
{$endif}

//***********************************************************************

{$ifdef BCB}
{$ObjExportAll On}
{$ASSERTIONS ON}
{$endif}


type
    EMemTableError = class(Exception);

    TRecInfo=record
        RecordNo: integer;
        BookmarkFlag: TBookmarkFlag;
    end;
    PRecInfo=^TRecInfo;

{
Internal buffer layout:
+------------------------+------------------------+---------------------------+
|     RECORD DATA        |    Rec.Information     |     Calculated Fields     |
| Record length bytes    |  SizeOf(TRecInfo) bytes|    CalcFieldSize bytes    |
+------------------------+------------------------+---------------------------+
                         ^                        ^
                    StartRecInfo              StartCalculated

Blobsfields in the internal buffer are pointers to the blob data.
}

  TkbmMemTableSaveFlag = (mtfSaveData, mtfSaveCalculated,mtfSaveLookup,mtfSaveNonVisible,mtfSaveBlobs,mtfSaveDef,mtfSkipRest);
  TkbmMemTableSaveFlags = set of TkbmMemTableSaveFlag;

  TkbmMemTableCompareOption = (mtcoDescending,mtcoCaseInsensitive,mtcoPartialKey);
  TkbmMemTableCompareOptions = set of TkbmMemTableCompareOption;

  TkbmBlob = class
  private
     FLength:cardinal;
     FBuffer:PChar;
  public
     constructor Create;
     destructor Destroy; override;

     procedure LoadFromStream(stream:TMemoryStream);
     procedure SaveToStream(stream:TMemoryStream);
     property Length:cardinal read FLength write FLength;
     property Buffer:Pchar read FBuffer write FBuffer;
  end;
  PkbmBlob=^TkbmBlob;

  TkbmOnLoadRecord = procedure(DataSet:TDataSet; var Accept:boolean) of object;
  TkbmOnLoadField = procedure(DataSet:TDataSet; FieldNo:integer; Field:TField) of object;
  TkbmOnCompress = procedure(UnCompressedStream,CompressedStream:TStream) of object;
  TkbmOnDeCompress = procedure(CompressedStream,DeCompressedStream:TStream) of object;

  TkbmMemTable = class(TDataSet)
  private
        FMasterLink:                            TMasterDataLink;
        FIsOpen:                                Boolean;
        FRecNo:                                 integer;
        FFilterBuffer:                          PChar;
        FRecords:                               TList;
        FAutoIncField:                          TField;
        FAutoIncMax,
        FBufferSize,
        FStartRecInfo,
        FStartCalculated:integer;
        FRecordSize:                            integer;
        FFieldOfs:                              array [0..255] of integer;
        FReadOnly:                              boolean;
        FPersistent:                            boolean;
        FPersistentFile:                        string;
        FBlobs:                                 boolean;
        FIndexList:                             TList;
        FSortList:                              TList;
        FRecalcOnIndex:                         boolean;
        FDummyStr:                              string;
        FIndexFieldNames:                       string;
        FSortFieldNames:                        string;
        FDoBinaryLocate:                        boolean;
        FSorted:                                boolean;

        FSortedOn:                              string;
        FSortOptions:                           TkbmMemTableCompareOptions;
        FKeyOptions:                            TkbmMemTableCompareOptions;

        FOnLoadRecord:                          TkbmOnLoadRecord;
        FOnLoadField:                           TkbmOnLoadField;

        FOnCompressSave:                        TkbmOnCompress;
        FOnDecompressLoad:                      TkbmOnDecompress;

        FOnCompressBlobStream:                  TkbmOnCompress;
        FOnDecompressBlobStream:                TkbmOnDecompress;

        FCapacity:                              integer;

        function GetActiveRecordBuffer:         PChar;

        function FilterRecord(Buffer: PChar):   Boolean;
        procedure _InternalAdd(Buffer:Pointer);
        procedure _InternalDelete(Pos:integer);
        procedure _InternalInsert(Pos:integer; Buffer:Pointer);
        procedure _InternalEmpty;
        procedure _InternalFirst;
        procedure _InternalLast;
        function  _InternalNext:boolean;
        function  _InternalPrior:boolean;
        procedure _InternalFreeRecord(RecPtr:Pointer);

        function  _InternalCompareRecords(FieldList:TList; Record1,Record2:PChar; Options:TkbmMemTableCompareOptions): Integer;
        function  _InternalKeyCompareRecords(FieldList:TList; KeyRecord,ARecord:PChar): Integer;
        function  _InternalSortCompareRecords(FieldList:TList; Record1,Record2:PChar): Integer;

        procedure _InternalQuickSort(sortlst:Tlist; L, R: Integer);
        function  _InternalBinarySearch(IndexLst:TList; First,Last:integer; KeyRecord:PChar; var Found:boolean):integer;
        function  _InternalSequentialSearch(IndexLst:TList; First,Last:integer; KeyRecord:PChar; var Found:boolean):integer;

        procedure SetMasterFields(const Value: string);
        function GetMasterFields: string;
        procedure SetDataSource(Value: TDataSource);

        function  IsFiltered:boolean;

        procedure BuildFieldList(List:TList; const FieldNames: string);
        procedure SetIndexFieldNames(const FieldNames:string);
        procedure SetSortFieldNames(const FieldNames:string);

        function GetVersion:string;
  protected
        procedure InternalOpen; override;
        procedure InternalClose; override;
        procedure InternalFirst;override;
        procedure InternalLast;override;

        procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
        procedure InternalDelete; override;
        procedure InternalInitRecord(Buffer: PChar); override;
        procedure InternalPost; override;

        procedure InternalInitFieldDefs; override;
        procedure InternalSetToRecord(Buffer: PChar); override;

        procedure InternalSort(sortlst:TList);

        procedure DoBeforeClose; override;
        procedure DoAfterOpen; override;
        procedure DoOnNewRecord; override;

        function IsCursorOpen: Boolean; override;
        function GetCanModify: Boolean; override;
        function GetRecordSize: Word;override;
        function GetRecordCount: integer;override;

        function AllocRecordBuffer: PChar; override;
        procedure FreeRecordBuffer(var Buffer: PChar); override;
        procedure PopulateField(Buffer:PChar;Field:TField;AValue:Variant);
        procedure PopulateRecord(Buffer:PChar;Fields:string;Values:variant);

        procedure CloseBlob(Field: TField); override;

        function GetFieldSize(FieldType:TFieldType; Size:longint):longint;
        function GetFieldPointer(Buffer:PChar; Field:TField):PChar;

        procedure SetFieldData(Field: TField; Buffer: Pointer);override;
{$IFDEF LEVEL3}
        function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
{$ENDIF}

        function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
        function FindRecord(Restart, GoForward: Boolean): Boolean; override;

        function GetRecNo: integer;override;
        procedure SetRecNo(Value: integer);override;

        function GetIsIndexField(Field: TField): Boolean; override;

        function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
        procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
        procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
        procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
        procedure InternalGotoBookmark(Bookmark: Pointer); override;

        function BCDToCurr(BCD: Pointer; var Curr: Currency): Boolean; override;
        function CurrToBCD(const Curr: Currency; BCD: Pointer; Precision, Decimals: Integer): Boolean; override;

        procedure InternalHandleException; override;

        procedure SetCapacity(const Capacity: integer);
        procedure SetCommaText(AString: String);
        function GetCommaText: String;
        function CopyRecords(Source,Destination:TDataSet; Count:longint):longint;
        procedure CopyFieldsProperties(Source,Destination:TDataSet);

        function LocateRecord(const KeyFields: string; const KeyValues: Variant;
                              Options: TLocateOptions; SyncCursor: Boolean): Boolean;

        function GetDataSource: TDataSource; override;
        function FilterMasterDetail:boolean;
        procedure MasterChanged(Sender: TObject);
        procedure MasterDisabled(Sender: TObject);
  public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;

{$IFDEF LEVEL4}
        function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
        procedure SetBlockReadSize(Value: Integer); override;
{$ENDIF}

        function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
        procedure CreateTable;
        procedure CreateTableAs(Source:TDataSet; CopyFieldProperties:boolean);
        procedure DeleteTable;

        procedure LoadFromFile(const FileName: string);
        procedure LoadFromStream(Stream:TStream);
        procedure LoadFromBinaryFile(const FileName: string);
        procedure LoadFromBinaryStream(Stream: TStream);
        procedure LoadFromDataSet(Source:TDataSet; CopyStructure,CopyFieldProperties:boolean);
        procedure SaveToFile(const FileName: string; flags:TkbmMemTableSaveFlags);
        procedure SaveToBinaryFile(const FileName: string; flags:TkbmMemTableSaveFlags);
        procedure SaveToStream(Stream: TStream; flags:TkbmMemTableSaveFlags);
        procedure SaveToBinaryStream(Stream: TStream; flags:TkbmMemTableSaveFlags);
        procedure SaveToDataSet(Destination:TDataSet);

        procedure EmptyTable;
        property CommaText:string read GetCommaText write SetCommaText;
        procedure Sort(Options:TkbmMemTableCompareOptions);
        procedure SortOn(const FieldNames:string; Options:TkbmMemTableCompareOptions);
        function IsSequenced:Boolean; override;
        function MoveRecord(Source, Destination: Integer): Boolean;
        function MoveCurRecord(Destination:Longint):Boolean;
        function Lookup(const KeyFields: string; const KeyValues: Variant; const ResultFields: string): Variant; override;
        function Locate(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions): Boolean; override;
        property Sorted:boolean read FSorted write FSorted stored false;

  published
        property Active;
        property Filtered;
        property IndexFields:string read FIndexFieldNames write SetIndexFieldNames;
        property SortFields:string read FSortFieldNames write SetSortFieldNames;
        property ReadOnly:boolean read FReadOnly write FReadOnly default false;
        property PersistentFile:string read FPersistentFile write FPersistentFile;
        property Persistent:boolean read FPersistent write FPersistent default false;
        property Filter:string read FDummyStr;
        property MasterFields: string read GetMasterFields write SetMasterFields;
        property MasterSource: TDataSource read GetDataSource write SetDataSource;
        property DoBinaryLocate:boolean read FDoBinaryLocate write FDoBinaryLocate;
        property Version:string read GetVersion write FDummyStr;
        property Capacity: integer read FCapacity write SetCapacity default 0;
        property OnLoadRecord:TkbmOnLoadRecord read FOnLoadRecord write FOnLoadRecord;
        property OnLoadField:TkbmOnLoadField read FOnLoadField write FOnLoadField;
        property OnCompressSave:TkbmOnCompress read FOnCompressSave write FOnCompressSave;
        property OnDecompressLoad:TkbmOnDecompress read FOnDecompressLoad write FOnDecompressLoad;
        property OnCompressBlobStream:TkbmOnCompress read FOnCompressBlobStream write FOnCompressBlobStream;
        property OnDecompressBlobStream:TkbmOnDecompress read FOnDecompressBlobStream write FOnDecompressBlobStream;
        property BeforeOpen;
        property AfterOpen;
        property BeforeClose;
        property AfterClose;
        property BeforeInsert;
        property AfterInsert;
        property BeforeEdit;
        property AfterEdit;
        property BeforePost;
        property AfterPost;
        property BeforeCancel;
        property AfterCancel;
        property BeforeDelete;
        property AfterDelete;
        property BeforeScroll;
        property AfterScroll;
        property OnCalcFields;
        property OnDeleteError;
        property OnEditError;
        property OnFilterRecord;
        property OnNewRecord;
        property OnPostError;
  end;

  TkbmBlobStream = class(TMemoryStream)
  private
    FField: TBlobField;
    FDataSet: TkbmMemTable;
    FMode:TBlobStreamMode;
    FPBlob: PkbmBlob;
    FPField: PChar;
    FFieldNo: Integer;
    FModified: Boolean;
    procedure ReadBlobData;
    procedure WriteBlobData;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    procedure Truncate;
  end;

procedure Register;

implementation

uses
  TypInfo, Dialogs, Windows, Forms, Controls, IniFiles;

const
  kbmSupportedFieldTypes=[ftString,ftSmallint,ftInteger,ftWord,ftBoolean,ftFloat,ftCurrency,
                          ftDate,ftTime,ftDateTime,ftAutoInc,ftBCD,ftBlob,ftMemo,ftGraphic,ftFmtMemo,
                          ftParadoxOle,ftDBaseOle,ftTypedBinary,ftBytes, ftVarBytes
{$ifndef LEVEL3}
                          ,ftFixedChar,ftWideString,ftLargeInt
{$endif}
                          ];

  kbmStringTypes=[ftString,ftMemo,ftFmtMemo
{$ifndef LEVEL3}
                          ,ftFixedChar,ftWideString
{$endif}
                          ];

  kbmBlobTypes=[ftBlob,ftMemo,ftGraphic,ftFmtMemo,ftParadoxOle,ftDBaseOle,ftTypedBinary,ftBytes,ftVarBytes];

{$ifdef LEVEL3}
  FieldTypeNames: array[TFieldType] of string = (
    'Unknown', 'String', 'SmallInt', 'Integer', 'Word', 'Boolean', 'Float',
    'Currency', 'BCD', 'Date', 'Time', 'DateTime', 'Bytes', 'VarBytes',
    'AutoInc', 'Blob', 'Memo', 'Graphic', 'FmtMemo', 'ParadoxOle',
    'dBaseOle', 'TypedBinary', 'Cursor');
{$endif}

  kbmNonBlobTypes=kbmSupportedFieldTypes - kbmBlobTypes;

  // Table definition magic words.
  kbmTableDefMagicStart = '@@TABLEDEF START@@';
  kbmTableDefMagicEnd = '@@TABLEDEF END@@';

  // Binary file magic word.
  kbmBinaryMagic = '@@BINARY@@';

// Compare two fields.
function CompareFields(p1,p2:pointer; FieldType: TFieldType; Partial, CaseInsensitive:Boolean):Integer;
var
   p:PChar;
   l:integer;
begin
     Result := 0;
     case FieldType of
{$ifndef LEVEL3}
       ftFixedChar,ftWideString,
{$endif}
       ftString:
          begin
               p:=nil;
               try
                  // If partial, cut to reference length. p1=reference field value, p2=tried field value.
                  if Partial then
                  begin
                       l:=StrLen(PChar(p1));
                       p:=StrAlloc(l+1);
                       StrLCopy(p,p2,l);
                  end
                  else p:=p2;

                  if CaseInsensitive then
                     Result:=AnsiCompareText(PChar(p1), p)
                  else
                     Result:=AnsiCompareStr(PChar(p1), p);
               finally
                  if p<>p2 then StrDispose(p);
               end;
          end;

       ftSmallint:
          Result:=SmallInt(p1^)-SmallInt(p2^);

       ftInteger,
       ftDate,
       ftTime,
       ftAutoInc:
          Result:=Longint(p1^)-Longint(p2^);

       ftWord:
          Result:=Word(p1^)-Word(p2^);

       ftBoolean:
          if WordBool(p1^)>WordBool(p2^) then Result:=1
          else if WordBool(p1^)<WordBool(p2^) then Result:=-1;

       ftFloat,
       ftCurrency:
          if Double(p1^)>Double(p2^) then Result:=1
          else if Double(p1^)<Double(p2^) then Result:=-1;

       ftDateTime:
          if TDateTime(p1^)>TDateTime(p2^) then Result:=1
          else if TDateTime(p1^)<TDateTime(p2^) then Result:=-1;
     end;
end;

// Code special characters (LF,CR,%,#0)
// CR (#13) -> %c
// LF (#10) -> %n
// #0 -> %0
// % -> %%
function StringToCodedString(const Source:string):string;
var
   i,j:integer;
   l:integer;
begin
     // Count CR/LF.
     l:=0;
     for i:=1 to length(Source) do
         if Source[i] in [#13,#10,'%',#0] then inc(l);

     // If no special characters, return the original string.
     if l=0 then
     begin
          Result:=Source;
          exit;
     end;

     // If any special characters, make room for them.
     SetLength(Result,length(Source)+l);

     // Code special characters.
     j:=1;
     for i:=1 to length(Source) do
         case Source[i] of
              #13: begin
                        Result[j]:='%'; inc(j);
                        Result[j]:='c'; inc(j);
                   end;
              #10: begin
                        Result[j]:='%'; inc(j);
                        Result[j]:='n'; inc(j);
                   end;
              #0:  begin
                        Result[j]:='%'; inc(j);
                        Result[j]:='0'; inc(j);
                   end;
              '%': begin
                        Result[j]:='%'; inc(j);
                        Result[j]:='%'; inc(j);
                   end;
              else
                   begin
                        Result[j]:=Source[i];
                        inc(j);
                   end;
         end;
end;

// Decode special characters (LF,CR,%,#0)
// %c -> CR (#13)
// %n -> LF (#10)
// %% -> %
// %0 -> #0
function CodedStringToString(const Source:string):string;
var
   i,j:integer;
begin
     SetLength(Result,length(Source));

     // Code special characters.
     i:=1;
     j:=1;
     while true do
     begin
          if i>length(Source) then break;
          if Source[i]='%' then
          begin
               inc(i);
               case Source[i] of
                    'c': Result[j]:=#13;
                    'n': Result[j]:=#10;
                    '%': Result[j]:='%';
                    '0': Result[j]:=#0;
               end;
               inc(j);
          end
          else
          begin
               Result[j]:=Source[i];
               inc(j);
          end;
          inc(i);
     end;

     // Cut result string to right length.
     if i<>j then SetLength(Result,j-1);
end;

// Code a string as BASE 64.
function StringToBase64(const Source:string):string;
var
   Act: Word;
   Bits,I,P,Len: Integer;
begin
     Bits:=0;
     Len:=(Length(Source)*4+2) div 3;
     if Len>0 then
     begin
          SetLength(Result,Len);
	  P:=1;
	  Act:=0;
	  for I:=1 to Pred(Len) do
          begin
	       if Bits<6 then
               begin
	            Act:=(Act shr 6) or (Ord(Source[P]) shl Bits);
		    Inc(P);
		    Inc(Bits,2);
               end
               else
               begin
	            Dec(Bits,6);
		    Act:=Act shr 6;
               end;
	       Result[I]:=Char(Act and 63+32);
          end;
	  Result[Len]:=Char(Act shr 6+32);
     end;
end;

// Decode BASE64 string.
function Base64ToString(const S: string): string;
var
   Act: Word;
   Bits,I,P,Len: Integer;
begin
     Len:=(Length(S)*3) div 4;
     SetLength(Result,Len);
     Bits:=0;
     Act:=0;
     P:=1;
     for I:=1 to system.Length(S) do
     begin
          Act:=Act or (Ord(S[I])-32) shl Bits;
	  if Bits>=2 then
          begin
	       Result[P]:=Char(Act and $FF);
	       Inc(P);
	       Act:=Act shr 8;
	       Dec(Bits,2);
          end
          else
	      Inc(Bits,6);
     end;
end;

// Compare two records.
function TkbmMemTable._InternalCompareRecords(FieldList:TList; Record1,Record2:PChar; Options:TkbmMemTableCompareOptions): Integer;
var
   p1,p2:PChar;
   fld:TField;
   i:integer;
begin
     Result:=0;

     // Loop through all indexfields, left to right.
     for i:=0 to FieldList.Count-1 do
     begin
          fld:=TField(FieldList[i]);

          // Get data for specified field for the two records.
          p1:=GetFieldPointer(Record1,fld);
          p2:=GetFieldPointer(Record2,fld);

          // Check if both not null.
          if (p1[0]<>#0) and (p2[0]<>#0) then
          begin
               // Skip null flag.
               inc(p1);
               inc(p2);

               // Compare the fields.
               Result:=CompareFields(p1,p2,fld.DataType,(mtcoPartialKey in Options),(mtcoCaseInsensitive in Options));
          end
          else if Boolean(p1[0]) then Result:=1
          else if Boolean(p2[0]) then Result:=-1;
          if Result<>0 then break;
     end;
end;

// Compare two arbitrary records for sort.
function TkbmMemTable._InternalSortCompareRecords(FieldList:TList; Record1,Record2:PChar): Integer;
begin
     // Check if to recalc before compare.
     if FRecalcOnIndex then
     begin
          //fill calc fields part of buffer
          ClearCalcFields(Record1);
          GetCalcFields(Record1);
          ClearCalcFields(Record2);
          GetCalcFields(Record2);
     end;

     // Compare record contents.
     Result:=_InternalCompareRecords(FieldList,Record1,Record2,FSortOptions);

     // Couldnt compare them according to fieldcontents, will now compare according to recnum.
     if Result=0 then
        Result:=PRecInfo(Record1+FStartRecInfo)^.RecordNo - PRecInfo(Record2+FStartRecInfo)^.RecordNo;

     // If descending sort, invert result.
     if (mtcoDescending in FSortOptions) then Result:=-Result;
end;

// Compare one arbitrary record with another fixed one for lookup.
function TkbmMemTable._InternalKeyCompareRecords(FieldList:TList; KeyRecord,ARecord:PChar): Integer;
begin
     // Check if to recalc before compare.
     if FRecalcOnIndex then
     begin
          //fill calc fields part of buffer
          ClearCalcFields(ARecord);
          GetCalcFields(ARecord);
     end;

     // Compare record contents.
     Result:=_InternalCompareRecords(FieldList,KeyRecord,ARecord,FKeyOptions);
end;

constructor TkbmMemTable.Create(AOwner: TComponent);
begin
     inherited Create(AOwner);
     FRecords:=TList.Create;
     FPersistent:=false;
     FRecalcOnIndex:=false;
     FIndexList:=TList.Create;
     FSortList:=TList.Create;

     FMasterLink:=TMasterDataLink.Create(Self);
     FMasterLink.OnMasterChange:=MasterChanged;
     FMasterLink.OnMasterDisable:=MasterDisabled;
end;

destructor TkbmMemTable.Destroy;
begin
     // Must be before deletion of records, otherwise it fails.
     inherited Destroy;

     // Delete allocated memory.
     FMasterLink.free;     FMasterLink:=nil;
     FIndexList.free;      FIndexList:=nil;
     FSortList.free;       FSortList:=nil;
     FRecords.free;        FRecords:=nil;
end;

// Get current component version.
function TkbmMemTable.GetVersion:string;
begin
     Result:=COMPONENT_VERSION;
end;

procedure TkbmMemTable.SetIndexFieldNames(const FieldNames:string);
begin
     try
        // Generate index fields.
        FIndexFieldNames:=FieldNames;
        BuildFieldList(FIndexList,FIndexFieldNames);
     except
        FIndexFieldNames:='';
        raise;
     end;
end;

procedure TkbmMemTable.SetSortFieldNames(const FieldNames:string);
begin
     try
        // Generate sort fields.
        FSortFieldNames:=FieldNames;
        BuildFieldList(FSortList,FSortFieldNames);
     except
        FSortFieldNames:='';
        raise;
     end;
end;

function TkbmMemTable.GetMasterFields: string;
begin
     Result:=FMasterLink.FieldNames;
end;

procedure TkbmMemTable.SetMasterFields(const Value: string);
begin
     FMasterLink.FieldNames:=Value;
end;

function TkbmMemTable.GetDataSource: TDataSource;
begin
     Result:=FMasterLink.DataSource;
end;

procedure TkbmMemTable.SetDataSource(Value: TDataSource);
begin
     if IsLinkedTo(Value) then DatabaseError('Selfreferencing master/detail relations not allowed.');
     FMasterLink.DataSource:=Value;
end;

procedure TkbmMemTable.MasterChanged(Sender: TObject);
begin
     CheckBrowseMode;
     First;
end;

procedure TkbmMemTable.MasterDisabled(Sender: TObject);
begin
     First;
end;

// Copy masterfields to detail table if a master/detail relation.
procedure TkbmMemTable.DoOnNewRecord;
var
   i:integer;
begin
     // Copy link values from master to detail.
     if FMasterLink.Active and (FMasterLink.Fields.Count > 0) then
        for i:=0 to FMasterLink.Fields.Count-1 do
            TField(FIndexList[i]).Value := TField(FMasterLink.Fields[i]).Value;

     inherited DoOnNewRecord;

     // Update autoinc if such a field is defined.
     if Assigned(FAutoIncField) then
        FAutoIncField.AsInteger:=FAutoIncMax+1;
end;

function TkbmMemTable.GetFieldSize(FieldType:TFieldType; Size:longint):longint;
begin
     case FieldType of
{$ifndef LEVEL3}
          ftFixedChar,ftWideString,
{$endif}
          ftString:             Result:=Size+1;
          ftSmallInt:           Result:=SizeOf(SmallInt);
          ftInteger:            Result:=SizeOf(Integer);
{$ifndef LEVEL3}
          ftLargeInt:           Result:=SizeOf(Int64);
{$endif}
          ftWord:               Result:=SizeOf(Word);
          ftBoolean:            Result:=SizeOf(WordBool);
          ftFloat:              Result:=SizeOf(Double);
          ftCurrency:           Result:=SizeOf(Double);
          ftDate:               Result:=SizeOf(Integer);
          ftTime:               Result:=SizeOf(Integer);
          ftDateTime:           Result:=SizeOf(Double);
          ftAutoInc:            Result:=SizeOf(Integer);
          ftBlob:               Result:=SizeOf(PkbmBlob);
          ftMemo:               Result:=SizeOf(PkbmBlob);
          ftGraphic:            Result:=SizeOf(PkbmBlob);
          ftFmtMemo:            Result:=SizeOf(PkbmBlob);
          ftParadoxOle:         Result:=SizeOf(PkbmBlob);
          ftDBaseOle:           Result:=SizeOf(PkbmBlob);
          ftTypedBinary:        Result:=SizeOf(PkbmBlob);
          ftBytes:              Result:=SizeOf(PkbmBlob);
          ftVarBytes:           Result:=SizeOf(PkbmBlob);
          ftBCD:                Result:=34;
     else
          Result:=0;
     end;
end;

function TkbmMemTable.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
begin
     Result:=TkbmBlobStream.Create(Field as TBlobField, Mode);
end;

procedure TkbmMemTable.CreateTable;
var
   i:Integer;
begin
     CheckInactive;

     // If no fielddefs existing, use the previously defined fields.
     if FieldDefs.Count = 0 then
        for i:=0 to FieldCount-1 do
            with Fields[i] do
                 if FieldKind = fkData then
                    FieldDefs.Add(FieldName, DataType, Size, Required);

     // Remove previously defined fields and create new from fielddefs.
     DestroyFields;
     CreateFields;
     FSorted:=false;
end;

// Create memory table as another dataset.
procedure TkbmMemTable.CreateTableAs(Source:TDataSet; CopyFieldProperties:boolean);
var
   i:integer;
begin
     CheckInactive;

     if Source=nil then exit;

     // Add fields as they are defined in the other dataset.
     Source.FieldDefs.Update;
     FieldDefs.Assign(Source.FieldDefs);

     // Remove non supported fieldsdefs.
     for i:=FieldDefs.Count-1 downto 0 do
         if not (FieldDefs.Items[i].DataType in kbmSupportedFieldTypes) then
            FieldDefs.Items[i].free;

     // Destroy existing fields.
     DestroyFields;
     if not Source.DefaultFields then CreateFields;

     // Copy fieldproperties from source.
     if CopyFieldProperties then CopyFieldsProperties(Source,self);
     FSorted:=false;
end;

// Delete table.
procedure TkbmMemTable.DeleteTable;
begin
     CheckInactive;
     DestroyFields;
end;

procedure TkbmMemTable._InternalAdd(Buffer:Pointer);
begin
     FRecords.Add(Buffer);
     FSorted:=false;
end;

procedure TkbmMemTable._InternalInsert(Pos:integer; Buffer:Pointer);
var
   i:integer;
   b:PChar;
begin
     if Pos<0 then Pos:=0;
     if (Pos = FRecords.Count) or(Pos = -1) then
        FRecords.Add(Buffer)
     else
        FRecords.Insert(Pos,Buffer);

     for i:=Pos+1 to FRecords.Count-1 do
     begin
          b:=FRecords.Items[i];
          inc(PRecInfo(b+FStartRecInfo).RecordNo);
     end;

     FSorted:=false;
end;

procedure TkbmMemTable._InternalFreeRecord(RecPtr:Pointer);
var
   i:integer;
   pblob:PkbmBlob;
begin
     if FBlobs then
     begin
          // Browse fields to delete blobs.
          for i:=0 to FieldCount-1 do
          begin
               if Fields[i].DataType in kbmBlobTypes then
               begin
                    pBlob:=PkbmBlob(GetFieldPointer(RecPtr,Fields[i])+1);
                    if pBlob^<>nil then
                    begin
                         pBlob^.free;
                         pBlob^:=nil;
                    end;
               end;
          end;
     end;
     FreeMem(RecPtr);
end;

procedure TkbmMemTable._InternalDelete(Pos:integer);
var
   i:integer;
   b:PChar;
begin
     _InternalFreeRecord(FRecords.Items[Pos]);
     FRecords.Delete(Pos);

     for i:=Pos to FRecords.Count-1 do
     begin
          b:=FRecords.Items[i];
          dec(PRecInfo(b+FStartRecInfo)^.RecordNo);
     end;
end;

// Purge all records.
procedure TkbmMemTable._InternalEmpty;
var
   i:integer;
begin
     for i:=0 to FRecords.Count-1 do
     begin
          _InternalFreeRecord(FRecords[i]);
     end;
     FRecords.Clear;
     FRecNo:=-1;
     FSorted:=false;
end;

procedure TkbmMemTable.InternalOpen;
var
   i: integer;
   sz: integer;
begin
     InternalInitFieldDefs;
     if DefaultFields then CreateFields;

     // Calculate field offsets into the record.
     FRecordSize:=0;
     FBlobs:=false;                  // Know of no blobs in the definition yet.
     for i:=0 to FieldDefs.Count - 1 do
         with FieldDefs[i] do
         begin
              // Check if field type supported.
              if not (DataType in kbmSupportedFieldTypes) then
                 raise EMemTableError.Create('Fielddef '+
                         Name
                         {$ifndef LEVEL3}
                         +' ('+DisplayName+')'
                         {$endif}
                         +' Datatype '
                         +inttostr(integer(DataType))
                         +' not supported.');
              FFieldOfs[i]:=FRecordSize;

              // Check if a blob.
              if (DataType in kbmBlobTypes) then FBlobs:=true;

              // Look for fieldsize.
              sz:=GetFieldSize(DataType,Size);
              inc(sz);   // 1.st byte is boolean flag for Null or not.
              inc(FRecordSize,sz);
         end;

     BindFields(True);
     FRecNo:=-1;
     BookmarkSize := sizeof(Integer);
     FStartRecInfo:=FRecordSize;
     FStartCalculated:=FStartRecInfo+SizeOf(TRecInfo);
     FBufferSize:=FRecordSize+Sizeof(TRecInfo)+CalcFieldsSize;
     FIsOpen:=True;
end;

procedure TkbmMemTable.InternalClose;
begin
     _InternalEmpty;
     FIsOpen:=False;
     BindFields(False);
     if DefaultFields then DestroyFields;
end;

procedure TkbmMemTable.InternalInitFieldDefs;
var
   i:integer;
begin
     FAutoIncField:=nil;
     FAutoIncMax:=-1;

     // If using predefined fields, generate fielddefs according to fields.
     if not DefaultFields then
     begin
          FieldDefs.clear;
          for i:=0 to Fieldcount-1 do
          begin
               FieldDefs.Add(Fields[i].FieldName,Fields[i].DataType,Fields[i].Size,Fields[i].Required);
          end;
     end;

     // Look for autoinc field if any.
     for i:=0 to FieldCount-1 do
         if Fields[i].DataType=ftAutoInc then
         begin
              FAutoIncField:=Fields[i];
              break;
         end;
end;

function TkbmMemTable.GetActiveRecordBuffer:  PChar;
begin
     case State of
          dsBrowse:        if IsEmpty then
                              Result := nil
                           else
                              Result := ActiveBuffer;
          dsCalcFields:    Result := CalcBuffer;
          dsFilter:        Result:=FFilterBuffer;
          dsEdit,dsInsert: Result:=ActiveBuffer;
          dsNewValue,dsOldValue,dsCurValue: Result:=ActiveBuffer;
{$IFDEF LEVEL4}
          dsBlockRead:     Result := ActiveBuffer;
{$ENDIF}
     else
          Result:=nil;
     end;
end;

function TkbmMemTable.GetFieldPointer(Buffer:PChar; Field:TField):PChar;
begin
     Result:=Buffer;
     if Buffer=nil then exit;
     if Field.FieldNo<=0 then
        inc(Result,FStartCalculated+Field.Offset)
     else
        inc(Result,FFieldOfs[Field.FieldNo-1]);
end;

// Result is data in the buffer and a boolean return (true=not null, false=is null).
function TkbmMemTable.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
   SourceBuffer: PChar;
begin
     Result:=False;
     if not FIsOpen then exit;
     SourceBuffer:=GetFieldPointer(GetActiveRecordBuffer,Field);
     if SourceBuffer=nil then Exit;

     if Assigned(Buffer) and (Field.DataType in kbmSupportedFieldTypes) then
          Move(SourceBuffer[1], Buffer^, GetFieldSize(Field.DataType,Field.Size));

     Result:=SourceBuffer[0]<>#0;
end;

procedure TkbmMemTable.SetFieldData(Field: TField; Buffer: Pointer);
var
   DestinationBuffer: PChar;
   sz:integer;
   l1,l2:integer;
begin
     if not FIsOpen then exit;
     DestinationBuffer:=GetFieldPointer(GetActiveRecordBuffer,Field);
     if DestinationBuffer=nil then Exit;

     if Field.ReadOnly then DatabaseErrorFmt('Field %s is read only',[Field.DisplayName]);
     if Field.FieldKind in [fkData, fkInternalCalc] then Field.Validate(Buffer);

     // Check if it could change the sort order.
     sz:=GetFieldSize(Field.DataType,Field.Size);
     if not ((Buffer<>nil) and (Boolean(DestinationBuffer[0]))) then FSorted:=false
     else if (Field.DataType in [{$ifndef LEVEL3}ftFixedChar,ftWideString,{$endif} ftString]) then
        begin
             l1:=strlen(Buffer);
             l2:=strlen(PChar(@DestinationBuffer[1]));
             if l1<>l2 then
                FSorted:=false
             else
                if not CompareMem(Buffer,PChar(@DestinationBuffer[1]),l1) then FSorted:=false;
        end
     else if not CompareMem(Buffer,PChar(@DestinationBuffer[1]),sz) then FSorted:=false;

     // Copy the data to the buffer.
     if LongBool(Buffer) then DestinationBuffer[0]:=#1
     else DestinationBuffer[0]:=#0;

     if Assigned(Buffer) and (Field.DataType in kbmSupportedFieldTypes) then
     begin
          if longbool(Buffer) then
             Move(Buffer^,DestinationBuffer[1],sz)
          else
             FillChar(DestinationBuffer[1],sz, 0);
     end;

     if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
        DataEvent(deFieldChange, Longint(Field));
end;

function TkbmMemTable.BCDToCurr(BCD: Pointer; var Curr: Currency): Boolean;
begin
     Move(BCD^, Curr, SizeOf(Currency));
     Result := True;
end;

function TkbmMemTable.CurrToBCD(const Curr: Currency; BCD: Pointer; Precision, Decimals: Integer): Boolean;
begin
     Move(Curr, BCD^, SizeOf(Currency));
     Result := True;
end;

function TkbmMemTable.IsCursorOpen: Boolean;
begin
     Result:=FIsOpen;
end;

function TkbmMemTable.GetCanModify: Boolean;
begin
     Result:=not FReadOnly;
end;

function TkbmMemTable.GetRecordSize: Word;
begin
     Result:=FRecordSize;
end;

function TkbmMemTable.AllocRecordBuffer: PChar;
begin
     GetMem(Result,FBufferSize);
     if Result<>nil then FillChar(Result^,FBufferSize,0);
end;

procedure TkbmMemTable.FreeRecordBuffer(var Buffer: PChar);
begin
     FreeMem(Buffer);
     Buffer:=nil;
end;

{$IFDEF LEVEL4}
procedure TkbmMemTable.SetBlockReadSize(Value: Integer);
var
   DoNext: Boolean;
begin
     if Value <> BlockReadSize then
     begin
          if (Value > 0) or (Value < -1) then
          begin
               inherited;
               BlockReadNext;
          end
          else
          begin
               DoNext:=Value=-1;
               Value:=0;
               inherited;

               if DoNext then
                  Next
               else
               begin
                    CursorPosChanged;
                    Resync([]);
               end;
          end;
     end;
end;
{$ENDIF}

// Fill one field with contents of a variant.
procedure TkbmMemTable.PopulateField(Buffer:PChar; Field:TField; AValue:Variant);
var
   p:PChar;
   pValue:PChar;

   s:array[0..dsMaxStringSize] of Char;
   si:smallint;
{$ifndef LEVEL3}
   li:LargeInt;
{$endif}
   i:integer;
   w:word;
   d:double;
   dt:TDateTime;
begin
     p:=GetFieldPointer(Buffer,Field);

     // Populate with null?
     if VarIsNull(AValue) then
     begin
          Boolean(p[0]):=false;
          FillChar(p[1],GetFieldSize(Field.DataType, Field.Size), 0);
          exit;
     end;

     // Populate with value.
     Boolean(p[0]):=true;
     with Field do
     begin
          pValue:=nil;
          case DataType of
{$ifndef LEVEL3}
               ftFixedChar,ftWideString,
{$endif}
               ftString:
                   begin
                        StrLCopy(s,PChar(VarToStr(AValue)),DataSize);
                        if TStringField(Field).Transliterate then
                           DataSet.Translate(s,s,True);
                        pValue:=s;
                   end;

               ftSmallint:
                   begin
                        si:=AValue;
                        pValue:=PChar(@si);
                   end;
{$ifndef LEVEL3}
               ftLargeInt:
                   begin
                        li:=trunc(double(AValue));
                        pValue:=PChar(@li);
                   end;
{$endif}
               ftInteger,
               ftDate,
               ftTime,
               ftAutoInc:
                   begin
                        i:=AValue;
                        pValue:=PChar(@i);
                   end;

               ftWord:
                   begin
                        w:=AValue;
                        pValue:=PChar(@w);
                   end;

               ftBoolean: ;

               ftFloat,
               ftCurrency:
                   begin
                        d:=AValue;
                        pValue:=PChar(@d);
                   end;

               ftDateTime:
                   begin
                        dt:=AValue;
                        pValue:=PChar(@dt);
                   end;
          end;
          if pValue<>nil then Move(pValue^,p[1],GetFieldSize(DataType,Size));
     end;
end;

// Fill record with values for specified fields.
procedure TkbmMemTable.PopulateRecord(Buffer:PChar; Fields:string; Values:variant);
var
   FieldList:Tlist;
   i:integer;
   n:integer;
begin
     FieldList := TList.Create;
     try
        GetFieldList(FieldList,Fields);

        n:=VarArrayDimCount(Values);
        if n>1 then raise EMemTableError.Create('Values variant array has invalid dimension count');
        if (n=0) and (FieldList.count>1) then raise EMemTableError.Create('More fields than values');
        if FieldList.Count<1 then raise EMemTableError.Create('There must be at least one field');

        // Single value.
        if n=0 then
        begin
             PopulateField(Buffer,TField(FieldList.Items[0]),Values);
             exit;
        end;

        // Several values.
        for i:=0 to FieldList.Count-1 do
        begin
             PopulateField(Buffer,TField(FieldList.Items[i]),Values[i]);
        end;
     finally
        FieldList.free;
     end;
end;

procedure TkbmMemTable.InternalFirst;
begin
     _InternalFirst;
end;

procedure TkbmMemTable.InternalLast;
begin
     _InternalLast;
end;

procedure TkbmMemTable._InternalFirst;
begin
     FRecNo:=-1;
end;

procedure TkbmMemTable._InternalLast;
begin
     FRecNo:=FRecords.Count;
end;

function TkbmMemTable._InternalNext:boolean;
begin
     if FrecNo<FRecords.Count-1 then
     begin
          Inc(FRecNo);
          Result:=true;
     end
     else Result:=false;
end;

function TkbmMemTable._InternalPrior:boolean;
begin
     if FrecNo>0 then
     begin
          Dec(FRecNo);
          Result:=true;
     end
     else Result:=false;
end;

function TkbmMemTable.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
   Acceptable: Boolean;
begin
     Result:=grOK;
     Acceptable:=False;
     repeat
         begin
              case GetMode of
                   gmCurrent: begin
                                   if FRecNo>=FRecords.Count then Result:=grEOF
                                   else if FRecNo<0 then Result:=grBOF
                                   else Result:=grOk;
                              end;
                   gmNext:    begin
                                   if _InternalNext then Result:=grOK
                                   else Result:=grEOF;
                              end;
                   gmPrior:   begin
                                   if _InternalPrior then Result:=grOK
                                   else Result:=grBOF;
                              end;
              end;
              if Result=grOk then
              begin
                      //fill TARrecord part of buffer
                      Move(FRecords.Items[FRecNo]^,Buffer^,FBufferSize);

                      //fill information part of buffer
                      with PRecInfo(Buffer+FStartRecInfo)^ do
                      begin
                           RecordNo:=FRecNo;
                           BookmarkFlag:=bfCurrent;
                      end;

                      //fill calc fields part of buffer
                      ClearCalcFields(Buffer);
                      GetCalcFields(Buffer);
                      Acceptable:=FilterRecord(Buffer);
                      if (GetMode=gmCurrent) and not Acceptable then Result:=grError;
              end
         end;
     until (Result<>grOk) or Acceptable;
end;

function TkbmMemTable.FindRecord(Restart, GoForward: Boolean): Boolean;
var
   Status:boolean;
begin
     CheckBrowseMode;
     DoBeforeScroll;
     SetFound(False);
     UpdateCursorPos;
     CursorPosChanged;

     if GoForward then
     begin
          if Restart then _InternalFirst;
          Status := _InternalNext;
     end else
     begin
          if Restart then _InternalLast;
          Status := _InternalPrior;
     end;

     if Status then
     begin
          Resync([rmExact, rmCenter]);
          SetFound(True);
     end;
     Result := Found;
     if Result then DoAfterScroll;
end;

// Is any record filtering applied.
// Could be master/detail or userdefined filter.
function TkbmMemTable.IsFiltered:boolean;
begin
     Result:=(Filtered and Assigned(OnFilterRecord)) or
             ((FMasterLink <> nil) and Assigned(FMasterLink.DataSource) and (FMasterLink.FieldNames<>'')
              and (FIndexList.Count>0));
end;

// Filter record according to master table.
function TkbmMemTable.FilterMasterDetail:boolean;
var
   i:integer;
begin
     Result:=false;

     // Check if to recalc before compare.
     if FRecalcOnIndex then
     begin
          ClearCalcFields(ActiveBuffer);
          GetCalcFields(ActiveBuffer);
     end;

     // Check if master detail defined but different from index fields. Then error.
     if FMasterLink.Fields.Count = 0 then exit;
     if FMasterLink.Fields.Count<>FIndexList.Count then
        DatabaseError('Number of masterfields doesnt correspond to number of indexfields.');

     // Check the fields.
     for i:=0 to FIndexList.Count-1 do
         if TField(FMasterLink.Fields.Items[i]).Value<>TField(FIndexList.Items[i]).Value then exit;
     Result:=true;
end;

// Filter records in general for
// master/detail and userdefined filter.
function TkbmMemTable.FilterRecord(Buffer: PChar): Boolean;
var
   SaveState: TDatasetState;
begin
     Result:=True;
     if not IsFiltered then Exit;

     // Now we will apply the filters on the record.
     SaveState:=SetTempState(dsFilter);
     FFilterBuffer:=Buffer;

     // Check if master/detail filtering.
     if Assigned(FMasterLink.DataSource) and (FMasterLink.FieldNames<>'') then
        Result:=FilterMasterDetail;

     // Call users own filtering if specified.
     if Result and Filtered and Assigned(OnFilterRecord) then OnFilterRecord(self,Result);

     // Finished filtering.
     RestoreState(SaveState);
end;

procedure TkbmMemTable.InternalSetToRecord(Buffer: PChar);
begin
     FRecNo:=PRecInfo(Buffer+FStartRecInfo).RecordNo;
end;

function TkbmMemTable.GetRecordCount: integer;
var
   SaveState: TDataSetState;
   SavePosition: integer;
   TempBuffer: PChar;
begin
     if not IsFiltered then Result:=FRecords.Count
     else
     begin
          Result:=0;
          SaveState:=SetTempState(dsBrowse);
          SavePosition:=FRecNo;
          try
             TempBuffer:=AllocRecordBuffer;
             InternalFirst;
             while GetRecord(TempBuffer,gmNext,True)=grOk do Inc(Result);
          finally
             RestoreState(SaveState);
             FRecNo:=SavePosition;
             FreeRecordBuffer(TempBuffer);
          end;
     end;
end;

function TkbmMemTable.GetRecNo: integer;
var
   SaveState: TDataSetState;
   SavePosition: integer;
   TempBuffer: PChar;
begin
     if not IsFiltered then
     begin
          UpdateCursorPos;
          if (FRecNo = -1) and (RecordCount > 0) then
             Result := 1
          else
             Result := FRecNo + 1;
     end
     else
     begin
          Result:=0;
          SaveState:=SetTempState(dsBrowse);
          SavePosition:=FRecNo;
          try
             TempBuffer:=AllocRecordBuffer;
             InternalFirst;
             if FRecords.Count > 0 then
             begin
                  while PRecInfo(TempBuffer+FStartRecInfo).RecordNo<>SavePosition do
                        if GetRecord(TempBuffer,gmNext,True)=grOk then Inc(Result);
             end;
          finally
             RestoreState(SaveState);
             FRecNo:=SavePosition;
             FreeRecordBuffer(TempBuffer);
          end;
     end;
end;

// Preallocate room for at least Capacity number of records in recordlist.
procedure TkbmMemTable.SetCapacity(const Capacity: integer);
begin
     if Capacity >= FRecords.Count then
     begin
          FCapacity := Capacity;
          FRecords.Capacity := Capacity;
     end;
end;

procedure TkbmMemTable.SetRecNo(Value: Integer);
var
   SaveState: TDataSetState;
   SavePosition: integer;
   TempBuffer: PChar;
begin
     if not IsFiltered then
     begin
          if (Value > 0) and (Value <= FRecords.Count) then
          begin
               FRecNo := Value - 1;
               {Resync([]);}
          end;
     end
     else
     begin
          SaveState:=SetTempState(dsBrowse);
          SavePosition:=FRecNo;
          try
             TempBuffer:=AllocRecordBuffer;
             InternalFirst;
             repeat
                   begin
                        if GetRecord(TempBuffer,gmNext,True)=grOk then Dec(Value)
                        else
                        begin
                             FRecNo:=SavePosition;
                             break;
                        end;
                   end;
             until Value=0;
          finally
             RestoreState(SaveState);
             FreeRecordBuffer(TempBuffer);
          end;
     end;

     // refresh the position of the dataset
     Resync([rmCenter,rmExact]);
end;

procedure TkbmMemTable.InternalAddRecord(Buffer: Pointer; Append: Boolean);
var
   b:Pointer;
begin
     // Allocate room for buffer in list.
     GetMem(b,FBufferSize);
     Move(Buffer^, b^, FBufferSize);
     if Append then
        _InternalAdd(b)
     else
         _InternalInsert(FRecNo,b);
end;

procedure TkbmMemTable.InternalDelete;
begin
     _InternalDelete(FRecNo);
end;

procedure TkbmMemTable.InternalInitRecord(Buffer: PChar);
begin
     FillChar(Buffer^,FBufferSize,0);
     PRecInfo(Buffer+FStartRecInfo)^.RecordNo:=FRecNo;
end;

procedure TkbmMemTable.InternalPost;
type
    TAutoInc=packed record
       Used:Byte;
       Int:Integer;
    end;
var
   b:pointer;
   n:integer;
   pAutoInc:^TAutoInc;
begin
     n:=PRecInfo(ActiveBuffer+FStartRecInfo)^.RecordNo;

     // If an autoinc field is specified, allways keey track of highest used number.
     if Assigned(FAutoIncField) then
     begin
          pAutoInc:=Pointer(GetFieldPointer(ActiveBuffer,FAutoIncField));
          if FAutoIncMax<pAutoInc^.Int then FAutoIncMax:=pAutoInc^.Int;
     end;

     if State = dsEdit then
        Move(ActiveBuffer^, FRecords.Items[n]^, FBufferSize)
     else
     begin
          GetMem(b,FBufferSize);
          Move(ActiveBuffer^, b^, FBufferSize);
          if GetBookmarkFlag(b) = bfEOF then
             _InternalAdd(b)
          else
             _InternalInsert(n,b);
     end;
end;

procedure TkbmMemTable.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
     PRecInfo(Buffer + FStartRecInfo).BookmarkFlag := Value;
end;

function TkbmMemTable.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
     Result:=PRecInfo(Buffer+FStartRecInfo).BookmarkFlag;
end;

procedure TkbmMemTable.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
     PInteger(Data)^ := PRecInfo(Buffer + FStartRecInfo).RecordNo;
end;

procedure TkbmMemTable.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
     PRecInfo(Buffer + FStartRecInfo).RecordNo := Pinteger(Data)^;
end;

procedure TkbmMemTable.InternalGotoBookmark (Bookmark: Pointer);
var
  ReqBookmark: Integer;
begin
     ReqBookmark := PInteger (Bookmark)^;
     if (ReqBookmark >= 0) and (ReqBookmark < FRecords.Count) then
        FRecNo := ReqBookmark
     else
        raise EMemTableError.Create('Bookmark ' + IntToStr(ReqBookmark) + ' not found');
end;

procedure TkbmMemTable.InternalHandleException;
begin
     Application.HandleException(Self);
end;

procedure TkbmMemTable.SaveToFile(const FileName: string; flags:TkbmMemTableSaveFlags);
var
   Stream: TStream;
begin
     Stream := TFileStream.Create(FileName, fmCreate);
     try
        SaveToStream(Stream,flags);
     finally
        Stream.Free;
     end;
end;

procedure TkbmMemTable.SaveToBinaryFile(const FileName: string; flags:TkbmMemTableSaveFlags);
var
   Stream: TStream;
begin
     Stream := TFileStream.Create(FileName, fmCreate);
     try
        SaveToBinaryStream(Stream,flags);
     finally
        Stream.Free;
     end;
end;

procedure TkbmMemTable.CloseBlob(Field: TField);
var
   pBlob:PkbmBlob;
begin
     if (FRecNo<0) or (FRecNo>=FRecords.Count) or (State<>dsEdit) then
     begin
          if Field.DataType in kbmBlobTypes then
          begin
               pBlob:=PkbmBlob(GetFieldPointer(ActiveBuffer,Field));
               if pBlob^<>nil then pBlob^:=nil;
          end;
     end;
end;

procedure TkbmMemTable.SaveToStream(Stream: TStream; flags:TkbmMemTableSaveFlags);
var
  i:integer;
  bm:TBookmark;
  nf:integer;
  s:string;
  l:integer;
  fset,f:^Boolean;
  Ods,Oms:char;
  oStream:TStream;
begin
  // Setup standard layout for data.
  Ods:=DateSeparator;
  Oms:=DecimalSeparator;
  DateSeparator:='/';
  DecimalSeparator:='.';

  bm:=GetBookmark;
  fset:=nil;
  oStream:=nil;
  try
     DisableControls;

     // If to compress stream, create memory stream to save to instead.
     if Assigned(FOnCompressSave) then
     begin
          oStream:=Stream;
          Stream:=TMemoryStream.Create;
     end;

     // Setup flags for fields to save.
     nf:=Fieldcount;
     GetMem(fset,nf * sizeof(boolean));
     f:=fset;
     for i:=0 to nf-1 do
     begin
          // Default dont save this field.
          f^:=false;

          // If a blob field, only save if specified.
          if (Fields[i].DataType in kbmBlobTypes) then
          begin
               if not (mtfSaveBlobs in flags) then continue;
               f^:=true;
          end;

          // Only save fields of specific types.
          case Fields[i].FieldKind of
               fkData: if mtfSaveData in flags then f^:=true;
               fkCalculated: if mtfSaveCalculated in flags then f^:=true;
               fkLookup: if mtfSaveLookup in flags then f^:=true;
               else f^:=true;
          end;

          // If not to save invisible fields, dont.
          if not (Fields[i].Visible or (mtfSaveNonVisible in flags)) then f^:=false;
          inc(f);
     end;

     // Write definition in CSV format.
     if mtfSaveDef in flags then
     begin
          // Write header.
          s:=AnsiQuotedStr(PChar(kbmTableDefMagicStart),'"')+#13+#10;
          l:=length(s);
          Stream.WriteBuffer(Pointer(s)^, l);

          // Write fielddefinitions.
          f:=fset;
          for i:=0 to nf-1 do
          begin
               if f^ or (not (mtfSkipRest in flags)) then
               begin
                    s:=Fields[i].FieldName+'='+
                        FieldTypeNames[Fields[i].DataType]+','+
                        inttostr(Fields[i].Size)+','+
                        AnsiQuotedStr(Fields[i].DisplayName,'"')+','+
                        AnsiQuotedStr(Fields[i].EditMask,'"')+','+
                        inttostr(Fields[i].DisplayWidth);
                    if Fields[i].Required then s:=s+',REQ';
                    if Fields[i].ReadOnly then s:=s+',RO';
                    s:=AnsiQuotedStr(PChar(s),'"')+#13+#10;
                    l:=length(s);
                    Stream.WriteBuffer(Pointer(s)^, l);
               end;

               inc(f);
          end;

          // Write footer.
          s:=AnsiQuotedStr(PChar(kbmTableDefMagicEnd),'"')+#13+#10;
          l:=length(s);
          Stream.WriteBuffer(Pointer(s)^, l);
     end;

     // Write all field display names in CSV format.
     s:='';
     f:=fset;
     for i:=0 to nf-1 do
     begin
          if f^ or (not (mtfSkipRest in flags)) then
             s:=s+AnsiQuotedStr(PChar(Fields[i].DisplayName),'"')+',';
          inc(f);
     end;
     s:=s+#13+#10;
     l:=length(s);
     Stream.Write(Pointer(s)^, l);

     // Write all records in CSV format.
     first;
     while not EOF do
     begin
          // Write current record.
          s:='';
          f:=fset;
          for i:=0 to nf-1 do
          begin
               if f^ then
               begin
                    if (Fields[i].IsNull) then s:=s+','
                    else if Fields[i].DataType in kbmStringTypes then
                        s:=s+AnsiQuotedStr(StringToCodedString(Fields[i].AsString),'"')+','
                    else if Fields[i].DataType in kbmBlobTypes then
                        s:=s+AnsiQuotedStr(StringToBase64(Fields[i].AsString),'"')+','
                    else
                        s:=s+AnsiQuotedStr(PChar(Fields[i].AsString),'"')+',';
               end
               else if not (mtfSkipRest in flags) then
                   s:=s+',';
               inc(f);
          end;

          // Next record.
          next;

          // Was it the last one?, then skip last ,.
          if EOF then System.Delete(s,length(s),1);
          s:=s+#13+#10;
          l:=length(s);

          // Write line.
          Stream.WriteBuffer(Pointer(s)^, l);
     end;

     // If to compress stream, send saved memorystream through event and save to original stream.
     if Assigned(FOnCompressSave) then FOnCompressSave(Stream,oStream);

  finally
     GotoBookmark(bm);
     EnableControls;
     FreeBookmark(bm);
     if fset<>nil then FreeMem(fset);
     if oStream<>nil then Stream.free;
     DateSeparator:=Ods;
     DecimalSeparator:=Oms;
  end;
end;

procedure TkbmMemTable.SaveToBinaryStream(Stream: TStream; flags:TkbmMemTableSaveFlags);
var
   i:integer;
   bm:TBookmark;
   nf:integer;
   s:string;
   fset,f:^Boolean;
   oStream:TStream;
   Writer : TWriter;
begin
     bm:=GetBookmark;
     fset:=nil;
     oStream:=nil;
     try
        DisableControls;

        // If to compress stream, create memory stream to save to instead.
        if Assigned(FOnCompressSave) then
        begin
             oStream:=Stream;
             Stream:=TMemoryStream.Create;
        end;

        Writer := TWriter.Create(Stream, 16384);
        Writer.WriteSignature;
        try

           // Setup flags for fields to save.
           nf:=Fieldcount;
           GetMem(fset,nf * sizeof(boolean));
           f:=fset;
           for i:=0 to nf-1 do
           begin
                // Default dont save this field.
                f^:=false;

                // If a blob field, only save if specified.
                if (Fields[i].DataType in kbmBlobTypes) then
                begin
                     if not (mtfSaveBlobs in flags) then continue;
                     f^:=true;
                end;

                // Only save fields of specific types.
                case Fields[i].FieldKind of
                     fkData: if mtfSaveData in flags then f^:=true;
                     fkCalculated: if mtfSaveCalculated in flags then f^:=true;
                     fkLookup: if mtfSaveLookup in flags then f^:=true;
                     else f^:=true;
                end;

                // If not to save invisible fields, dont.
                if not (Fields[i].Visible or (mtfSaveNonVisible in flags)) then f^:=false;
                inc(f);
           end;

           // Write field definitions
           Writer.WriteListBegin;

           // Write fielddefinitions.
           f:=fset;
           for i:=0 to nf-1 do
           begin
                if f^ then
                begin
                     Writer.WriteString(Fields[i].FieldName);
                     Writer.WriteString(FieldTypeNames[Fields[i].DataType]);
                     Writer.WriteInteger(Fields[i].Size);
                     Writer.WriteString(Fields[i].DisplayName);
                     Writer.WriteString(Fields[i].EditMask);
                     Writer.WriteInteger(Fields[i].DisplayWidth);
                     Writer.WriteBoolean(Fields[i].Required);
                     Writer.WriteBoolean(Fields[i].ReadOnly);
                end;

                inc(f);
           end;
           Writer.WriteListEnd;

           // Write all records
           Writer.WriteListBegin;
           first;
           while not EOF do
           begin
                // Write current record.
                s:='';
                f:=fset;
                for i:=0 to nf-1 do
                begin
                     if f^ then
                     begin
                          case Fields[i].DataType of
                               ftBoolean : Writer.WriteBoolean(Fields[i].AsBoolean);

   {$ifndef LEVEL3}
                               ftLargeInt,
   {$endif}
                               ftSmallInt,
                               ftInteger,
                               ftWord,
                               ftAutoInc : Writer.WriteInteger(Fields[i].AsInteger);

                               ftFloat : Writer.WriteFloat(Fields[i].AsFloat);

                               ftBCD,
                               ftCurrency : Writer.WriteFloat(Fields[i].AsCurrency);

                               ftDate,
                               ftTime,ftDateTime: Writer.WriteFloat(Fields[i].AsFloat);
                          else
                              Writer.WriteString(Fields[i].AsString);
                          end;
                     end;

                     inc(f);
                end;

                // Next record.
                next;
           end;
           Writer.WriteListEnd;

        finally
           Writer.FlushBuffer;
           Writer.Free;

           // If to compress stream, send saved memorystream through event and save to original stream.
           if Assigned(FOnCompressSave) then FOnCompressSave(Stream,oStream);
        end;

     finally
        GotoBookmark(bm);
        EnableControls;
        FreeBookmark(bm);
        if oStream<>nil then Stream.free;
        if fset<>nil then FreeMem(fset);
     end;
end;

procedure TkbmMemTable.LoadFromFile(const FileName: string);
var
   Stream: TStream;
begin
     Stream := TFileStream.Create(FileName, fmOpenRead);
     try
        LoadFromStream(Stream);
     finally
        Stream.Free;
     end;
end;

procedure TkbmMemTable.LoadFromBinaryFile(const FileName: string);
var
   Stream: TStream;
begin
     Stream := TFileStream.Create(FileName, fmOpenRead);
     try
        LoadFromBinaryStream(Stream)
     finally
        Stream.Free;
     end;
end;

procedure TkbmMemTable.LoadFromStream(Stream: TStream);
const
   BUFSIZE=8192;
var
   i:integer;
   bm:TBookmark;
   nf:integer;
   s:string;
   buf,bufptr:PChar;
   remaining_in_buf:integer;
   Line,Word:string;
   lptr,elptr:PChar;
   null:boolean;
   Ods,Oms:char;
   DuringTableDef:boolean;
   slist:TStringList;
   FName,TName,DName,EMask:string;
   FSize,DSize:integer;
   REQ,RO:boolean;
   FT:TFieldType;
   Accept:boolean;
   oStream:TStream;

   function GetChunk:boolean;
   begin
        remaining_in_buf:=Stream.Read(pointer(buf)^,BUFSIZE);
        bufptr:=buf;
        Result:=remaining_in_buf>0;
   end;

   function GetLine:boolean;
   var
     EOL,EOF:boolean;
     ep,sp:PChar;
     TmpStr:string;
   begin
        // Cut out a line.
        EOL:=false;
        EOF:=false;
        Line:='';
        sp:=bufptr;
        ep:=bufptr;
        while true do
        begin
             // Check if need another chunk.
             if remaining_in_buf=0 then
             begin
                  // Add to line.
                  SetString(TmpStr,sp,bufptr-sp);
                  Line:=Line+TmpStr;

                  // Check if EOF.
                  if not GetChunk then
                  begin
                       EOF:=true;
                       break;
                  end;
                  sp:=bufptr;
             end;

             // Check if we got EOL character, skip them and finally break.
             if (bufptr^) in [#0, #10, #13] then
             begin
                  if not EOL then ep:=bufptr-1;
                  EOL:=true
             end
             else if EOL then
             begin
                  SetString(TmpStr,sp,ep-sp+1);
                  Line:=Line+TmpStr;
                  break;
             end;

             // Prepare to look at next char.
             Inc(bufptr);
             dec(remaining_in_buf);
        end;

        lptr:=PChar(Line);
        elptr:=PChar(Line)+Length(Line)-1;
        Result:=(not EOF);
   end;

   function GetWord(var null:boolean):string;
   label
     L_exit;
   var
     sptr:PChar;
     TmpStr:string;
   begin

     // Cut out next word.
     Result:='';

     // Look for starting " or ,.
     while (lptr^ <> '"') and (lptr^ <> ',') and (lptr<elptr) do inc(lptr);
     if (lptr>=elptr) then exit;
     if (lptr^ = ',') then
     begin
          null:=true;
          inc(lptr);
          exit;
     end
     else null:=false;
     inc(lptr);

     while true do
     begin
          // Look for ending ".
          sptr:=lptr;
          while not (lptr^ = '"') do
          begin
               if (lptr>=elptr) then goto L_exit;
               inc(lptr);
          end;
          SetString(TmpStr,sptr,lptr-sptr);
          Result:=Result+TmpStr;
          inc(lptr);

          // Is it a double "" or end of word ?.
          if (lptr^ = '"') then
          begin
               Result:=Result+'"';
               inc(lptr);
               continue;
          end;

L_exit:
          inc(lptr);
          break;
     end;
   end;

begin
  // Setup standard layout for data.
  Ods:=DateSeparator;
  Oms:=DecimalSeparator;
  DateSeparator:='/';
  DecimalSeparator:='.';

  bm:=GetBookmark;

  oStream:=nil;
  try
     // Allocate space for a buffer.
     GetMem(buf,BUFSIZE);

     // Still nothing in the buffer to handle.
     remaining_in_buf:=0;

     DisableControls;

     // If to decompress stream, create memory stream to load from instead.
     if Assigned(FOnDeCompressLoad) then
     begin
          oStream:=Stream;
          Stream:=TMemoryStream.Create;
          FOnDecompressLoad(oStream,Stream);
          Stream.Position:=0;
     end;

     // Read all definition lines in CSV format.
     slist:=TStringList.Create;
     DuringTableDef:=false;
     try
        while true do
        begin
             GetLine;
             if Line='' then break;

             // Start of table definition?
             Word:=GetWord(null);
             if Word=kbmTableDefMagicStart then
             begin
                  DuringTableDef:=true;
                  Close;
                  FieldDefs.clear;
                  DeleteTable;
                  continue;
             end

             // End of table definition?
             else if Word=kbmTableDefMagicEnd then
             begin
                  DuringTableDef:=false;
                  Open;
                  continue;
             end;

             // If not during table definitions then its the header. Break.
             if not DuringTableDef then break;

             // Otherwise its a field definition. Break the line apart.
             i:=pos('=',Word);
             slist.CommaText:=copy(Word,i+1,length(Line));
             FName:=copy(Word,1,i-1);
             TName:=slist.Strings[0];
             FSize:=strtoint(slist.Strings[1]);
             DName:=slist.Strings[2];
             EMask:=slist.Strings[3];
             DSize:=strtoint(slist.Strings[4]);
             REQ:=pos(',REQ',Word)<>0;
             RO:=pos(',RO',Word)<>0;

             // Find fieldtype from fieldtypename.
             for i:=0 to ord(High(FieldTypeNames)) do
                 if FieldTypeNames[TFieldType(i)]=TName then break;
             FT:=TFieldType(i);
             if not (FT in kbmSupportedFieldTypes) then
                raise EMemTableError.Create('Unknown field type ('+TName+') in CSV file. ('+Word+')');

             // Add field definition.
             FieldDefs.Add(FName,FT,FSize,REQ);

             // Setup other properties.
             i:=FieldDefs.IndexOf(FName);
             with FieldDefs.Items[i].CreateField(self) do
             begin
                  DisplayLabel:=DName;
                  EditMask:=EMask;
                  ReadOnly:=RO;
                  DisplayWidth:=DSize;
             end;
        end;
     finally
        slist.free;
     end;

     // Read all lines in CSV format.
     nf:=Fieldcount;
     while true do
     begin
          GetLine;
          if Line='' then break;

          append;

          i:=0;
          while (lptr<elptr) and (i<nf) do
          begin
               s:=GetWord(null);
               if null then
                   Fields[i].Clear
               else if Fields[i].DataType in kbmStringTypes then
                   Fields[i].AsString:=CodedStringToString(s)
               else if Fields[i].DataType in kbmBlobTypes then
                   Fields[i].AsString:=Base64ToString(s)
               else
                   Fields[i].AsString:=s;

               if Assigned(FOnLoadField) then FOnLoadField(self,i,Fields[i]);
               inc(i);
          end;

          Accept:=true;
          if Assigned(FOnLoadRecord) then FOnLoadRecord(self,Accept);
          if Accept then Post
          else Cancel;
     end;
  finally
     FreeMem(buf);
     if oStream<>nil then Stream.free;
     GotoBookmark(bm);
     EnableControls;
     FreeBookmark(bm);
     DateSeparator:=Ods;
     DecimalSeparator:=Oms;
  end;
end;

procedure TkbmMemTable.LoadFromBinaryStream(Stream: TStream);
var
   i:integer;
   bm:TBookmark;
   nf:integer;
   FName,TName,DName,EMask:string;
   FSize,DSize:integer;
   REQ,RO:boolean;
   FT:TFieldType;
   Accept:boolean;
   oStream:TStream;
   Reader : TReader;
begin
     bm:=GetBookmark;

     oStream:=nil;
     Reader := nil;
     try
        DisableControls;

        // If to decompress stream, create memory stream to load from instead.
        if Assigned(FOnDeCompressLoad) then
        begin
             oStream:=Stream;
             Stream:=TMemoryStream.Create;
             FOnDecompressLoad(oStream,Stream);
             Stream.Position:=0;
        end;

        Reader := TReader.Create(Stream, 16384);
        Reader.ReadSignature;

        // Read all definitions if any saved.
        try
           Reader.ReadListBegin;
           Close;
           FieldDefs.clear;
           DeleteTable;

           while not(Reader.EndofList) do
           begin
                // read field definition.
                FName := Reader.ReadString;
                TName := Reader.ReadString;
                FSize := Reader.ReadInteger;
                DName := Reader.ReadString;
                EMask := Reader.ReadString;
                DSize := Reader.ReadInteger;
                REQ := Reader.ReadBoolean;
                RO := Reader.ReadBoolean;

                // Find fieldtype from fieldtypename.
                for i:=0 to ord(High(FieldTypeNames)) do
                    if FieldTypeNames[TFieldType(i)]=TName then break;

                FT:=TFieldType(i);

                if not (FT in kbmSupportedFieldTypes) then
                   raise EMemTableError.Create('Unknown field type ('+TName+') in file.');

                // Add field definition.
                FieldDefs.Add(FName,FT,FSize,REQ);

                // Setup other properties.
                i:=FieldDefs.IndexOf(FName);
                with FieldDefs.Items[i].CreateField(self) do
                begin
                     DisplayLabel:=DName;
                     EditMask:=EMask;
                     ReadOnly:=RO;
                     DisplayWidth:=DSize;
                end;
           end;
           Reader.ReadListEnd;
        finally
           Open;
        end;

        // Read data.
        nf:=Fieldcount;
        Reader.ReadListBegin;
        while not(Reader.EndofList) do
        begin
             append;
             for i:=0 to nf-1 do
             begin
                  case Fields[i].DataType of
                       ftBoolean : Fields[i].AsBoolean := Reader.ReadBoolean;

{$ifndef LEVEL3}
                       ftLargeInt,
{$endif}
                       ftSmallInt,
                       ftInteger,
                       ftWord,
                       ftAutoInc : Fields[i].AsInteger := Reader.ReadInteger;

                       ftFloat : Fields[i].AsFloat := Reader.ReadFloat;

                       ftBCD,
                       ftCurrency : Fields[i].AsCurrency := Reader.ReadFloat;

                       ftDate,
                       ftTime,
                       ftDateTime : Fields[i].AsFloat := Reader.ReadFloat;
                  else
                      Fields[i].AsString := Reader.ReadString;
                  end;

                  if Assigned(FOnLoadField) then FOnLoadField(self,i,Fields[i]);
             end;

             Accept:=true;
             if Assigned(FOnLoadRecord) then FOnLoadRecord(self,Accept);
             if Accept then Post
             else Cancel;
        end;
        Reader.ReadListEnd;
     finally
        if Assigned(Reader) then reader.Free;
        if oStream<>nil then Stream.free;
        GotoBookmark(bm);
        EnableControls;
        FreeBookmark(bm);
     end;
end;

procedure TkbmMemTable.EmptyTable;
begin
     DisableControls;
     try
        _InternalEmpty;
        First;
     finally
         EnableControls;
     end;
end;

procedure TkbmMemTable.SetCommaText(AString: String);
var
   stream:TMemoryStream;
begin
     EmptyTable;
     stream:=TMemoryStream.Create;
     try
        stream.Write(Pointer(AString)^,length(AString));
        stream.Seek(0,soFromBeginning);
        LoadFromStream(stream);
     finally
        stream.free;
     end;
end;

function TkbmMemTable.GetCommaText: String;
var
   stream:TMemoryStream;
   sz:integer;
   p:PChar;
begin
     Result:='';
     stream:=TMemoryStream.Create;
     try
        SaveToStream(stream,[mtfSaveData]);
        stream.Seek(0,soFromBeginning);
        sz:=stream.Size;
        p:=stream.Memory;
        setstring(Result,p,sz);
     finally
        stream.free;
     end;
end;

// Sneak in before the table is closed.
procedure TkbmMemTable.DoBeforeClose;
begin
     inherited;

     // If persistent, save info to file.
     if FPersistent and (FPersistentFile <> '') then
     begin
          try
             SysUtils.DeleteFile(FPersistentFile);
          finally
             SaveToFile(FPersistentFile,[mtfSaveData,mtfSaveNonVisible]);
          end;
     end;
end;

// Sneak in after the table is opened.
procedure TkbmMemTable.DoAfterOpen;
begin
     // If persistent, read info from file.
     if FPersistent and (FPersistentFile <> '') and FileExists(FPersistentFile) then
     begin
          LoadFromFile(FPersistentFile);
          first;
     end;

     inherited;
end;

// Binary search routine.
// Needs the data to be sorted on the indexfields defined by indexlst beforehand.
// Recursive function.
function TkbmMemTable._InternalBinarySearch(IndexLst:TList; First,Last:integer; KeyRecord:PChar; var Found:boolean):integer;
var
   Mid:integer;
begin
     // Look in the center of the interval.
     Mid:=(Last+First) div 2;

     // Check key record equal to record in middle of area.
     Result:=_InternalKeyCompareRecords(IndexLst,KeyRecord,FRecords.List^[Mid]);
     if Result=0 then
     begin
          if IsFiltered then
          begin
               // Look sequentially for matching unfiltered record.
               repeat
                     if FilterRecord(FRecords.List^[Mid]) then
                     begin
                          Found:=true;
                          break;
                     end;
                     inc(Mid);
               until (Mid>Last) or (_InternalKeyCompareRecords(IndexLst,KeyRecord,FRecords.List^[MId]) <> 0);
          end
          else
              Found:=true;
          if Found then
          begin
               RecNo:=Mid+1;
               UpdateCursorPos;
               Resync([rmCenter,rmExact]);
          end;
          exit;
     end;

     // Not matching, dig deeper.
     Found:=false;

     // If the interval is just one record, nothing more to search.
     if First=Last then
     begin
          Result:=0;
          exit;
     end;

     // Check if sorted in descending order, reverse result.
     if mtcoDescending in FSortOptions then Result:=-Result;

     // If the key is smaller than the middle record, look in the lower half segment.
     if Result<0 then
        Result:=_InternalBinarySearch(IndexLst,First,Mid-1,KeyRecord,Found)

     // If the key is bigger than the middle record, look in the upper half segment.
     else if Result>0 then
        Result:=_InternalBinarySearch(IndexLst,Mid+1,Last,KeyRecord,Found);
end;

// Sequential search.
function TkbmMemTable._InternalSequentialSearch(IndexLst:TList; First,Last:integer; KeyRecord:PChar; var Found:boolean):integer;
var
   i:integer;
   f:boolean;
begin
     // Loop for all records.
     f:=IsFiltered;
     Result:=1;
     for i:=0 to FRecords.Count-1 do
     begin
          // Check filtering.
          if f and not FilterRecord(FRecords.List^[i]) then continue;

          // Check key record equal to record.
          Result:=_InternalKeyCompareRecords(IndexLst,KeyRecord,FRecords.List^[i]);
          if Result=0 then
          begin
               Found:=true;
               RecNo:=i+1;
               UpdateCursorPos;
               Resync([rmCenter,rmExact]);
               exit;
          end;
     end;
end;

// Locate record.
// If the keyfields are the same as sorted fields and the table is currently sorted,
// it will make a fast binary search. Otherwise it will make a sequential search.
// Binary searches dont take partial record in account.
function TkbmMemTable.LocateRecord(const KeyFields: string;
                                   const KeyValues: Variant; Options: TLocateOptions;
                                   SyncCursor: Boolean): Boolean;
var
   i:integer;
   KeyFieldsList: TList;
   Found:boolean;
   KeyRecord:PChar;
begin
     // Setup key options.
     FKeyOptions:=[];
     if loCaseInsensitive in Options then FKeyOptions:=FkeyOptions+[mtcoCaseInsensitive];
     if loPartialKey in Options then FKeyOptions:=FKeyOptions+[mtcoPartialKey];

     Result := False;
     I := VarArrayDimCount(KeyValues);
     if I > 1 then
        raise EMemTableError.Create('KeyValues VarArray invalid DimCount');
     CheckBrowseMode;
     CursorPosChanged;

     // Prepare list of fields representing the keys to search for.
     KeyFieldsList := TList.Create;
     try
        GetFieldList(KeyFieldsList, KeyFields);

        // Populate a keyrecord.
        KeyRecord:=AllocRecordBuffer;
        try
           // Fill it with values.
           PopulateRecord(KeyRecord,KeyFields,KeyValues);

           // Check if to recalc reference record before compare.
           if FRecalcOnIndex then
           begin
                //fill calc fields part of buffer
                ClearCalcFields(KeyRecord);
                GetCalcFields(KeyRecord);
           end;

           DisableControls;
           try
              Found:=false;

              // Make fast binary search when the data are ordered by keyFields...
              if FDoBinaryLocate and FSorted and (not IsFiltered) and (pos(KeyFields,FSortedOn)=1) then
              begin
                   // Do binary locate.
                   _InternalBinarySearch(KeyFieldsList,0,FRecords.Count-1,KeyRecord,Found);
                   exit;
              end;

              // Otherwise do a sequential search.
              _InternalSequentialSearch(KeyFieldsList,0,FRecords.Count-1,KeyRecord,Found);
           finally
              EnableControls;
           end;

        finally
           // Free reference record.
           FreeRecordBuffer(KeyRecord);
        end;

     finally
        KeyFieldsList.Free;
        SetFound(Found);
        Result:=Found;
     end;
end;

function TkbmMemTable.Lookup(const KeyFields: string; const KeyValues: Variant; const ResultFields: string): Variant;
var
   BM: TBookmarkStr;
begin
     Result := Null;
     BM := Bookmark;
     try
        if LocateRecord(KeyFields, KeyValues, [], False) then
        begin
             SetTempState(dsCalcFields);
             try
                CalculateFields(ActiveBuffer);
                Result := FieldValues[ResultFields];
             finally
                RestoreState(dsBrowse);
             end;
        end;
     finally
        Bookmark := BM;
     end;
end;

function TkbmMemTable.Locate(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions): Boolean;
begin
     DoBeforeScroll;
     Result := LocateRecord(KeyFields, KeyValues, Options, True);
     if Result then
     begin
          //Resync([rmExact, rmCenter]);
          DoAfterScroll;
     end;
end;

// Copy properties from source to destination.
// Handles different fieldorder between the two datasets.
procedure TkbmMemTable.CopyFieldsProperties(Source,Destination:TDataSet);
var
   i:integer;
   fc:integer;
   f:TField;
begin
     // Did we get valid parameters.
     if (Source=nil) or (Destination=nil) or (Source=Destination) then exit;

     // Copy constraints from source to destination.
     fc:=Destination.FieldCount-1;
     for i:=0 to fc do
     begin
          // Find matching fieldnames on both sides. If fieldname not found, dont copy it.
          f:=Source.FindField(Destination.Fields[i].FieldName);
          if f=nil then continue;

          // Copy general properties.
          Fields[i].EditMask:=f.EditMask;
          Fields[i].DisplayWidth:=f.DisplayWidth;
          Fields[i].DisplayLabel:=f.DisplayLabel;
          Fields[i].Required:=f.Required;
          Fields[i].ReadOnly:=f.ReadOnly;
          Fields[i].Visible:=f.Visible;
          Fields[i].DefaultExpression:=f.DefaultExpression;
          Fields[i].Alignment:=f.Alignment;

          // Copy field type specific properties.
          if f is TNumericField then
             with TNumericField(f) do
             begin
                  TNumericField(Fields[i]).DisplayFormat:=DisplayFormat;
                  TNumericField(Fields[i]).EditFormat:=EditFormat;
             end;

          if f is TIntegerField then
             with TIntegerField(f) do
             begin
                  TIntegerField(Fields[i]).MaxValue:=MaxValue;
                  TIntegerField(Fields[i]).MinValue:=MinValue;
             end;

          if f is TDateTimeField then
             with TDateTimeField(f) do
                  TDateTimeField(Fields[i]).DisplayFormat:=DisplayFormat;

          if f is TBooleanField then
             with TBooleanField(f) do
                  TBooleanField(Fields[i]).DisplayValues:=DisplayValues;

          if f is TStringField then
             with TStringField(f) do
                  TStringField(Fields[i]).Transliterate:=Transliterate;

          if f is TFloatField then
             with TFloatField(f) do
             begin
                  TFloatField(Fields[i]).MaxValue:=MaxValue;
                  TFloatField(Fields[i]).MinValue:=MinValue;
                  TFloatField(Fields[i]).Precision:=Precision;
                  TFloatField(Fields[i]).currency:=currency;
             end;

          if f is TBCDField then
             with TBCDField(f) do
             begin
                  TBCDField(Fields[i]).MaxValue:=MaxValue;
                  TBCDField(Fields[i]).MinValue:=MinValue;
{$IFDEF LEVEL4}
                  TBCDField(Fields[i]).Precision:=Precision;
{$ENDIF}
                  TBCDField(Fields[i]).currency:=currency;
             end;

          if f is TBlobField then
             with TBlobField(f) do
             begin
                  TBlobField(Fields[i]).BlobType:=BlobType;
                  TBlobField(Fields[i]).Transliterate:=Transliterate;
             end;
     end;
end;

// Copy records from source to destination.
// Handles different fieldorder between the two datasets.
// Returns the number of records copied.
function TkbmMemTable.CopyRecords(Source,Destination:TDataSet;Count:longint):longint;
var
   i:integer;
   fc:integer;
   f:TField;
   fi:array [0..255] of integer;
   Accept:boolean;
begin
     Result:=0;

     // Did we get valid parameters.
     if (Source=nil) or (Destination=nil) or (Source=Destination) then exit;

     // Build name index relations between destination and source dataset.
     fc:=Destination.FieldCount-1;
     for i:=0 to fc do
     begin
          // Check if not a datafield or not a supported field, dont copy it.
          if (Destination.Fields[i].FieldKind<>fkData) or
             (not (Destination.Fields[i].DataType in kbmSupportedFieldTypes)) then
          begin
               fi[i]:=-1;
               continue;
          end;

          // Find matching fieldnames on both sides. If fieldname not found, dont copy it.
          f:=Source.FindField(Destination.Fields[i].FieldName);
          if f=nil then
          begin
               fi[i]:=-1;
               continue;
          end;

          fi[i]:=f.Index;
     end;

     // Copy data.
     Source.First;
     while not Source.EOF do
     begin
          Destination.Append;
          for i:=0 to fc do
          begin
               if fi[i]>=0 then
                  Destination.Fields[i].Value:=Source.Fields[fi[i]].Value
               else
                  Destination.Fields[i].Clear;
               if Assigned(FOnLoadField) and (Destination=self) then FOnLoadField(Self,i,Destination.Fields[i]);
          end;

          Accept:=true;
          if Assigned(FOnLoadRecord) and (Destination=self) then FOnLoadRecord(Self,Accept);
          if Accept then Destination.post
          else Destination.Cancel;

          Source.next;
          inc(Result);
          if (Count>0) and (Result>Count) then break;
     end;
end;

// Fill the memorytable with data from another dataset.
procedure TkbmMemTable.LoadFromDataSet(Source:TDataSet; CopyStructure,CopyFieldProperties:boolean);
var
   SourceActive:boolean;
   SourceDisabled:boolean;
   OldMasterSource:TDataSource;
   OldFiltered:boolean;
   BM:TBookmark;
begin
     if Source=self then exit;

     // Close this table.
     Close;

     // Remember state of source.
     SourceActive:=Source.Active;
     SourceDisabled:=Source.ControlsDisabled;

     // Dont update controls while scrolling through source.
     Source.DisableControls;

     // Remember state of this.
     OldFiltered:=Filtered;
     OldMasterSource:=MasterSource;

     try
        DisableControls;
        Source.Open;
        BM:=Source.GetBookmark;
        try

           // Dont want to check filtering while copying.
           Filtered := False;
           MasterSource:=nil;

           // Open source.
           Source.CheckBrowseMode;
           Source.UpdateCursorPos;

           // Create this memorytable as a copy of the other one.
           if CopyStructure then CreateTableAs(Source,CopyFieldProperties);
           Open;
           CheckBrowseMode;

           // Move to first record in source.
           Source.First;
           CopyRecords(Source,self,-1);
           Source.GotoBookmark(BM);
        finally
           First;
           Filtered:=OldFiltered;
           MasterSource:=OldMasterSource;
           FreeBookmark(BM);
        end;
     finally
        EnableControls;
        if not SourceActive then Source.Close;
        if not SourceDisabled then Source.EnableControls;
        UpdateCursorPos;
        CursorPosChanged;
     end;
end;

// Append the data in this memory table to another dataset.
procedure TkbmMemTable.SaveToDataSet(Destination:TDataSet);
var
   DestActive:boolean;
   DestDisabled:boolean;
begin
     if Destination=self then exit;

     // Remember state of destination.
     DestActive:=Destination.Active;
     DestDisabled:=Destination.ControlsDisabled;

     // Dont update controls while appending to destination
     Destination.DisableControls;

     try
        DisableControls;
        try
           // Open destination
           Destination.Open;
           Destination.CheckBrowseMode;
           Destination.UpdateCursorPos;

           // Open this if not opened.
           Open;
           CheckBrowseMode;

           // Move to first record in this.
           First;
           CopyRecords(self,Destination,-1);
        finally
           Destination.First;
        end;
     finally
        EnableControls;
        if not DestActive then Destination.Close;
        if not DestDisabled then Destination.EnableControls;
     end;
end;

function TkbmMemTable.IsSequenced: Boolean;
begin
     Result:=not Filtered;
end;

// Record rearranging.

// Move record from one place in table to another.
function TkbmMemTable.MoveRecord(Source, Destination: Integer): Boolean;
var
   p: Pointer;
   b: PChar;
   I: Integer;
begin
     Result := False;
     {Because property RecNo has values 1..FRecords.Count
      and FRecNo has values 0..FRecords.Count - 1}
     Dec(Source);
     Dec(Destination);

     if (Source <> Destination) and (Source > -1) and (Source < FRecords.Count)
        and (Destination > -1) and (Destination < FRecords.Count) then
     begin
          p := FRecords[Source];
          FRecords.Delete(Source);
          for i:=Source to FRecords.Count-1 do
          begin
               b:=FRecords.Items[i];
               dec(PRecInfo(b+FStartRecInfo)^.RecordNo);
          end;
          _InternalInsert(Destination,p);
          RecNo := Destination + 1;
     end;
end;

// Move record to the specified destination.
function TkbmMemTable.MoveCurRecord(Destination: Integer): Boolean;
begin
     Result := MoveRecord(RecNo,Destination);
end;

// Sorting.

// Quick sort algorithm for our purpose.
procedure TkbmMemTable._InternalQuickSort(sortlst:TList; L, R: Integer);
var
   I, J: Integer;
   P, T: Pointer;
begin
     repeat
           I:=L;
           J:=R;
           P:=FRecords.List^[(L + R) shr 1];
           repeat
                 while _InternalSortCompareRecords(sortlst,FRecords.List^[I],P) < 0 do Inc(I);
                 while _InternalSortCompareRecords(sortlst,FRecords.List^[J],P) > 0 do Dec(J);
                 if I <= J then
                 begin
                      T := FRecords.List^[I];
                      FRecords.List^[I]:=FRecords.List^[J];
                      FRecords.List^[J]:=T;
                      Inc(I);
                      Dec(J);
                 end;
           until I>J;
           if L<J then _InternalQuickSort(sortlst,L,J);
           L:=I;
    until I>=R;
end;

// Callback function for TDataset to know if specified field is an index.
function TkbmMemTable.GetIsIndexField(Field:TField):Boolean;
begin
     Result:=FIndexList.IndexOf(Field)>=0;
end;

// Build field list from list of fieldnames.
procedure TkbmMemTable.BuildFieldList(List:TList; const FieldNames: string);
var
   p:integer;
   fld:TField;
begin
     List.Clear;
     p:=1;
     FRecalcOnIndex:=false;
     while p<=length(FieldNames) do
     begin
          fld:=FieldByName(ExtractFieldName(FieldNames,p));
          if ((fld.FieldKind=fkData) or (fld.FieldKind=fkCalculated)) and (fld.DataType in (kbmSupportedFieldTypes-kbmBlobTypes)) then
             List.Add(fld)
          else
              DatabaseErrorFmt('Cant index on field %s',[fld.DisplayName]);
          if fld.FieldKind=fkCalculated then FRecalcOnIndex:=true;
     end;
end;

// Internal sort function.
procedure TkbmMemTable.InternalSort(sortlst:TList);
var
   Pos: TBookmarkStr;
begin
     if (not Active) or (FRecords.Count<=0) then exit;

     if sortlst.Count<=0 then exit;

     // Remember our position.
     Pos:=Bookmark;
     try
        _InternalQuickSort(sortlst,0,FRecords.Count-1);
        SetBufListSize(0);
        try
           SetBufListSize(BufferCount + 1);
        except
           SetState(dsInactive);
           CloseCursor;
           raise;
        end;
     finally
        // Reposition.
        Bookmark:=Pos;
     end;
     FSorted:=true;
     Resync([]);
end;

// Do sort on specified sortfields.
procedure TkbmMemTable.Sort(Options:TkbmMemTableCompareOptions);
begin
     FSortOptions:=Options;
     FSortedOn:=SortFields;
     InternalSort(FSortList);
end;

// Do sort on specifed fieldnames.
procedure TkbmMemTable.SortOn(const FieldNames:string; Options:TkbmMemTableCompareOptions);
var
   lst:Tlist;
begin
     FSortOptions:=Options;
     FSortedOn:=FieldNames;
     lst:=Tlist.create;
     try
        BuildFieldList(lst,FieldNames);
        InternalSort(lst);
     finally
        lst.free;
     end;
end;

// TkbmBlob.

constructor TkbmBlob.Create;
begin
     inherited;
     FLength:=0;
     FBuffer:=nil;
end;

destructor TkbmBlob.Destroy;
begin
     FLength:=0;
     if FBuffer<>nil then
     begin
          FreeMem(FBuffer);
          FBuffer:=nil;
     end;
     inherited;
end;

procedure TkbmBlob.LoadFromStream(stream:TMemoryStream);
begin
     FLength:=stream.Size;
     FBuffer:=Allocmem(FLength);
     stream.Seek(0,soFromBeginning);
     stream.Read(FBuffer^,FLength);
end;

procedure TkbmBlob.SaveToStream(stream:TMemoryStream);
begin
     stream.SetSize(FLength);
     stream.Seek(0,soFromBeginning);
     stream.Write(FBuffer^,FLength);
     stream.Seek(0,soFromBeginning);
end;

// TkbmBlobStream

// On create, make a stream access to the specified blobfield in the current record.
constructor TkbmBlobStream.Create(Field:TBlobField;Mode:TBlobStreamMode);
begin
     FField:=Field;
     FFieldNo:=FField.FieldNo;
     FDataSet:=FField.DataSet as TkbmMemTable;
     FMode:=Mode;
     FPField:=FDataSet.GetFieldPointer(FDataSet.GetActiveRecordBuffer,Field);
     if FPField=nil then Exit;
     FPBlob:=PkbmBlob(FPField+1);
     if Mode<>bmRead then
     begin
          if FField.ReadOnly then DatabaseErrorFmt('Field %s is readonly.',[FField.DisplayName]);
          if not (FDataSet.State in [dsEdit, dsInsert]) then DatabaseError('Dataset is not in edit mode.');
     end;
     if Mode=bmWrite then
          Truncate
     else
        ReadBlobData;
end;

// On destroy, update the blobfield in the current record if the blob has changed.
destructor TkbmBlobStream.Destroy;
begin
     if FModified then
     try
        WriteBlobData;
        FField.Modified:=true;

        FDataSet.DataEvent(deFieldChange,Longint(FField));
     except
        Application.HandleException(Self);
     end;
     inherited Destroy;
end;

procedure TkbmBlobStream.WriteBlobData;
var
   blob:TkbmBlob;
   Stream:TMemoryStream;
begin
     // Get old allocation if any, and free it.
     if FPBlob^<>nil then
     begin
          FPBlob^.free;
          FPBlob^:=nil;
     end;

     // Make new blob with data and enter it into the field
     blob:=TkbmBlob.Create;
     FPBlob^:=blob;

     // If to compress the blob data, do it.
     if Assigned(FDataSet.FOnCompressBlobStream) then
     begin
          Stream:=TMemoryStream.Create;
          try
             FDataSet.FOnCompressBlobStream(self,Stream);
             Stream.Position:=0;
             blob.LoadFromStream(Stream);
          finally
             Stream.free;
          end;
     end
     else
         // Otherwise just save raw data to the inmemory blob.
         blob.LoadFromStream(self);

     // Set Null flag in record.
     if Size<>0 then FPField[0]:=#1
     else FPField[0]:=#0;
end;

procedure TkbmBlobStream.ReadBlobData;
var
   blob:TkbmBlob;
   Stream:TMemoryStream;
begin
     // Get allocation.
     blob:=FPBlob^;
     if blob=nil then exit;

     // If to decompress stream, save the blob in a memory stream and decompress it.
     if Assigned(FDataSet.FOnDeCompressBlobStream) then
     begin
          Stream:=TMemoryStream.Create;
          try
             blob.SaveToStream(Stream);
             FDataSet.FOnDecompressBlobStream(Stream,self);
          finally
             Stream.free;
          end;
     end
     else
         // Copy the data to the stream.
         blob.SaveToStream(self);
     self.Position:=0;
end;

function TkbmBlobStream.Write(const Buffer;Count:Longint): Longint;
begin
     Result:=inherited Write(Buffer,Count);
     if FMode=bmWrite then FModified:=true;
end;

procedure TkbmBlobStream.Truncate;
begin
     Clear;

     // If blob allocated, remove allocation.
     if FPBlob^<>nil then
     begin
          FPBlob^.free;
          FPBlob^:=nil;
     end;

     FModified:=true;
end;

procedure Register;
begin
     RegisterComponents('Data Access', [TkbmMemTable]);
end;

end.
